from __future__ import unicode_literals

from django.utils import six
from djblets.features.testing import override_feature_checks
from djblets.webapi.errors import DOES_NOT_EXIST, INVALID_FORM_DATA
from djblets.webapi.testing.decorators import webapi_test_template

from reviewboard.changedescs.models import ChangeDescription
from reviewboard.reviews.models.status_update import StatusUpdate
from reviewboard.webapi.resources import resources
from reviewboard.webapi.tests.base import BaseWebAPITestCase
from reviewboard.webapi.tests.mimetypes import (status_update_item_mimetype,
                                                status_update_list_mimetype)
from reviewboard.webapi.tests.mixins import (BasicTestsMetaclass,
                                             ReviewRequestChildItemMixin,
                                             ReviewRequestChildListMixin)
from reviewboard.webapi.tests.mixins_extra_data import (ExtraDataItemMixin,
                                                        ExtraDataListMixin)
from reviewboard.webapi.tests.urls import (get_review_item_url,
                                           get_status_update_item_url,
                                           get_status_update_list_url)


@six.add_metaclass(BasicTestsMetaclass)
class ResourceListTests(ExtraDataListMixin, ReviewRequestChildListMixin,
                        BaseWebAPITestCase):
    """Testing the StatusUpdateResource list APIs."""

    fixtures = ['test_users']
    resource = resources.status_update

    sample_api_url = '/review-requests/<id>/status-updates/'

    def setup_review_request_child_test(self, review_request):
        return (get_status_update_list_url(review_request),
                status_update_list_mimetype)

    def compare_item(self, item_rsp, status_update):
        self.assertEqual(item_rsp['id'], status_update.pk)
        self.assertEqual(item_rsp['summary'], status_update.summary)
        self.assertEqual(item_rsp['description'], status_update.description)
        self.assertEqual(item_rsp['url'], status_update.url)
        self.assertEqual(item_rsp['url_text'], status_update.url_text)
        self.assertEqual(item_rsp['extra_data'], status_update.extra_data)

    #
    # HTTP GET tests
    #

    def setup_basic_get_test(self, user, with_local_site, local_site_name,
                             populate_items):
        review_request = self.create_review_request(
            with_local_site=with_local_site,
            submitter=user,
            publish=True)

        if populate_items:
            items = [self.create_status_update(review_request)]
        else:
            items = []

        return (get_status_update_list_url(review_request, local_site_name),
                status_update_list_mimetype,
                items)

    def test_get_with_change_id(self):
        """Testing the GET /review-requests/<id>/status-updates/?change= API"""
        review_request = self.create_review_request(
            submitter=self.user,
            publish=True)

        change = ChangeDescription.objects.create()
        review_request.changedescs.add(change)

        self.create_status_update(review_request)
        update = self.create_status_update(review_request,
                                           change_description=change)

        with override_feature_checks(self.override_features):
            rsp = self.api_get(
                get_status_update_list_url(review_request),
                {'change': change.pk},
                expected_mimetype=status_update_list_mimetype)

        self.assertEqual(rsp['stat'], 'ok')
        self.assertEqual(len(rsp['status_updates']), 1)
        self.assertEqual(rsp['status_updates'][0]['id'], update.pk)

    def test_get_with_service_id(self):
        """Testing the GET /review-requests/<id>/status-updates/?service-id=
        API"""
        review_request = self.create_review_request(
            submitter=self.user,
            publish=True)

        self.create_status_update(review_request, service_id='service1')
        update = self.create_status_update(review_request,
                                           service_id='service2')

        with override_feature_checks(self.override_features):
            rsp = self.api_get(
                get_status_update_list_url(review_request),
                {'service-id': update.service_id},
                expected_mimetype=status_update_list_mimetype)

        self.assertEqual(rsp['stat'], 'ok')
        self.assertEqual(len(rsp['status_updates']), 1)
        self.assertEqual(rsp['status_updates'][0]['id'], update.pk)

    def test_get_with_state(self):
        """Testing the GET /review-requests/<id>/status-updates/?state= API"""
        review_request = self.create_review_request(
            submitter=self.user,
            publish=True)

        self.create_status_update(review_request, state=StatusUpdate.PENDING)
        update = self.create_status_update(review_request,
                                           state=StatusUpdate.DONE_SUCCESS)

        with override_feature_checks(self.override_features):
            rsp = self.api_get(
                get_status_update_list_url(review_request),
                {'state': 'done-success'},
                expected_mimetype=status_update_list_mimetype)

        self.assertEqual(rsp['stat'], 'ok')
        self.assertEqual(len(rsp['status_updates']), 1)
        self.assertEqual(rsp['status_updates'][0]['id'], update.pk)

    #
    # HTTP POST tests
    #

    def setup_basic_post_test(self, user, with_local_site, local_site_name,
                              post_valid_data):
        review_request = self.create_review_request(
            with_local_site=with_local_site,
            submitter=user,
            publish=True)

        if post_valid_data:
            post_data = {
                'service_id': 'Service',
                'summary': 'Summary',
            }
        else:
            post_data = {}

        return (
            get_status_update_list_url(review_request, local_site_name),
            status_update_item_mimetype,
            post_data,
            [review_request])

    def check_post_result(self, user, rsp, review_request):
        status_update = StatusUpdate.objects.get(pk=rsp['status_update']['id'])
        self.compare_item(rsp['status_update'], status_update)

    @webapi_test_template
    def test_post_with_invalid_state(self):
        """Testing the POST <URL> API with an invalid state"""
        review_request = self.create_review_request(publish=True)

        with override_feature_checks(self.override_features):
            rsp = self.api_post(
                get_status_update_list_url(review_request),
                {
                    'service_id': 'Service',
                    'summary': 'Summary',
                    'state': 'incorrect',
                },
                expected_status=400)

        self.assertEqual(rsp['stat'], 'fail')
        self.assertEqual(rsp['err']['code'], INVALID_FORM_DATA.code)
        self.assertTrue('state' in rsp['fields'])

    @webapi_test_template
    def test_post_with_invalid_change_id(self):
        """Testing the POST <URL> API with an change_id state"""
        review_request = self.create_review_request(publish=True)

        with override_feature_checks(self.override_features):
            rsp = self.api_post(
                get_status_update_list_url(review_request),
                {
                    'service_id': 'Service',
                    'summary': 'Summary',
                    'change_id': '123456',
                },
                expected_status=400)

        self.assertEqual(rsp['stat'], 'fail')
        self.assertEqual(rsp['err']['code'], INVALID_FORM_DATA.code)
        self.assertTrue('change_id' in rsp['fields'])

    @webapi_test_template
    def test_post_with_invalid_review_id(self):
        """Testing the POST <URL> API with an invalid review_id"""
        review_request = self.create_review_request(publish=True)

        with override_feature_checks(self.override_features):
            rsp = self.api_post(
                get_status_update_list_url(review_request),
                {
                    'service_id': 'Service',
                    'summary': 'Summary',
                    'review_id': '123456',
                },
                expected_status=400)

        self.assertEqual(rsp['stat'], 'fail')
        self.assertEqual(rsp['err']['code'], INVALID_FORM_DATA.code)
        self.assertTrue('review_id' in rsp['fields'])


@six.add_metaclass(BasicTestsMetaclass)
class ResourceItemTests(ReviewRequestChildItemMixin, ExtraDataItemMixin,
                        BaseWebAPITestCase):
    """Testing the StatusUpdateResource item APIs."""

    fixtures = ['test_users']
    sample_api_url = 'review-requests/<id>/status-updates/<id>/'
    resource = resources.status_update

    def setup_review_request_child_test(self, review_request):
        status_update = self.create_status_update(review_request)

        return (get_status_update_item_url(review_request, status_update.pk),
                status_update_item_mimetype)

    def compare_item(self, item_rsp, status_update):
        self.assertEqual(item_rsp['id'], status_update.pk)

        if status_update.review_id:
            review_request = status_update.review_request
            local_site_name = (review_request.local_site and
                               review_request.local_site.name)

            review_url = self.base_url + get_review_item_url(
                review_request, status_update.review_id, local_site_name)

            self.assertEqual(item_rsp['links']['review']['href'], review_url)

    #
    # HTTP DELETE tests
    #

    def setup_basic_delete_test(self, user, with_local_site, local_site_name):
        review_request = self.create_review_request(
            with_local_site=with_local_site,
            submitter=user,
            publish=True)
        status_update = self.create_status_update(review_request, user=user)

        return (get_status_update_item_url(review_request, status_update.pk,
                                           local_site_name),
                [status_update, review_request])

    def check_delete_result(self, user, status_update, review_request):
        self.assertNotIn(status_update, review_request.status_updates.all())

    @webapi_test_template
    def test_delete_with_does_not_exist(self):
        """Testing the DELETE <URL> API
        with Does Not Exist error
        """
        review_request = self.create_review_request(publish=True)

        with override_feature_checks(self.override_features):
            rsp = self.api_delete(
                get_status_update_item_url(review_request, 12345),
                expected_status=404)

        self.assertEqual(rsp['stat'], 'fail')
        self.assertEqual(rsp['err']['code'], DOES_NOT_EXIST.code)

    #
    # HTTP GET tests
    #

    def setup_basic_get_test(self, user, with_local_site, local_site_name):
        review_request = self.create_review_request(
            with_local_site=with_local_site,
            submitter=user,
            publish=True)
        status_update = self.create_status_update(review_request, user=user)

        return (get_status_update_item_url(review_request, status_update.pk,
                                           local_site_name),
                status_update_item_mimetype,
                status_update)

    @webapi_test_template
    def test_get_not_modified(self):
        """Testing the GET <URL> API
        with Not Modified response
        """
        review_request = self.create_review_request(publish=True)
        status_update = self.create_status_update(review_request)

        with override_feature_checks(self.override_features):
            self._testHttpCaching(
                get_status_update_item_url(review_request, status_update.pk),
                check_etags=True)

    #
    # HTTP PUT tests
    #

    def setup_basic_put_test(self, user, with_local_site, local_site_name,
                             put_valid_data):
        review_request = self.create_review_request(
            with_local_site=with_local_site,
            submitter=user,
            publish=True)
        review = self.create_review(review_request=review_request)
        status_update = self.create_status_update(review_request, user=user)

        return (
            get_status_update_item_url(review_request, status_update.pk,
                                       local_site_name),
            status_update_item_mimetype,
            {
                'summary': 'New summary',
                'review_id': review.pk,
            },
            status_update,
            [])

    def check_put_result(self, user, item_rsp, status_update, *args):
        self.assertEqual(item_rsp['id'], status_update.pk)

        status_update = StatusUpdate.objects.get(pk=status_update.pk)
        self.compare_item(item_rsp, status_update)



from __future__ import unicode_literals

from django.contrib.auth.models import User
from django.http import Http404
from django.test.client import RequestFactory
from django.test.utils import override_settings

from reviewboard.notifications.email.message import EmailMessage
from reviewboard.notifications.email.views import BasePreviewEmailView
from reviewboard.testing import TestCase


class BasePreviewEmailViewTests(TestCase):
    """Unit tests for BasePreviewEmailView."""

    @override_settings(DEBUG=True)
    def test_get_with_classmethod(self):
        """Testing BasePreviewEmailView.get with build_email as classmethod"""
        class MyPreviewEmailView(BasePreviewEmailView):
            @classmethod
            def build_email(cls, test_var):
                self.assertEqual(test_var, 'test')
                return EmailMessage(subject='Test Subject',
                                    text_body='Test Body')

            def get_email_data(view, request, test_var=None, *args, **kwargs):
                self.assertEqual(test_var, 'test')

                return {
                    'test_var': test_var,
                }

        request = RequestFactory().request()
        request.user = User.objects.create_user(username='test-user',
                                                email='user@example.com')

        view = MyPreviewEmailView.as_view()
        response = view(request, test_var='test', message_format='text')

        self.assertEqual(response.status_code, 200)

    @override_settings(DEBUG=True)
    def test_get_with_staticmethod(self):
        """Testing BasePreviewEmailView.get with build_email as staticmethod"""
        class MyPreviewEmailView(BasePreviewEmailView):
            @staticmethod
            def build_email(test_var):
                self.assertEqual(test_var, 'test')
                return EmailMessage(subject='Test Subject',
                                    text_body='Test Body')

            def get_email_data(view, request, test_var=None, *args, **kwargs):
                self.assertEqual(test_var, 'test')

                return {
                    'test_var': test_var,
                }

        request = RequestFactory().request()
        request.user = User.objects.create_user(username='test-user',
                                                email='user@example.com')

        view = MyPreviewEmailView.as_view()
        response = view(request, test_var='test', message_format='text')

        self.assertEqual(response.status_code, 200)

    @override_settings(DEBUG=False)
    def test_get_with_debug_false(self):
        """Testing BasePreviewEmailView.get with DEBUG=False"""
        class MyPreviewEmailView(BasePreviewEmailView):
            @classmethod
            def build_email(cls, test_var):
                self.fail('build_email should not be reached')
                return EmailMessage(subject='Test Subject',
                                    text_body='Test Body')

            def get_email_data(view, request, test_var=None, *args, **kwargs):
                self.fail('get_email_data should not be reached')

        request = RequestFactory().request()
        request.user = User.objects.create_user(username='test-user',
                                                email='user@example.com')

        view = MyPreviewEmailView.as_view()

        with self.assertRaises(Http404):
            view(request, test_var='test', message_format='text')



"""Unit tests for the Assembla hosting service."""

from __future__ import unicode_literals

import nose

from reviewboard.admin.server import get_hostname
from reviewboard.hostingsvcs.testing import HostingServiceTestCase
from reviewboard.scmtools.crypto_utils import encrypt_password
from reviewboard.scmtools.models import Repository, Tool


class AssemblaTestCase(HostingServiceTestCase):
    """Base class for Assembla test suites."""

    service_name = 'assembla'
    fixtures = ['test_scmtools']

    default_account_data = {
        'password': encrypt_password('abc123'),
    }


class AssemblaTests(AssemblaTestCase):
    """Unit tests for the Assembla hosting service."""

    def test_service_support(self):
        """Testing Assembla service support capabilities"""
        self.assertTrue(self.service_class.needs_authorization)
        self.assertTrue(self.service_class.supports_bug_trackers)
        self.assertTrue(self.service_class.supports_repositories)
        self.assertEqual(self.service_class.supported_scmtools,
                         ['Perforce', 'Subversion'])

    def test_get_repository_fields_with_perforce(self):
        """Testing Assembla.get_repository_fields for Perforce"""
        self.assertEqual(
            self.get_repository_fields(
                'Perforce',
                fields={
                    'assembla_project_id': 'myproject',
                }
            ),
            {
                'path': 'perforce.assembla.com:1666',
                'encoding': 'utf8',
            })

    def test_get_repository_fields_with_subversion(self):
        """Testing Assembla.get_repository_fields for Subversion"""
        self.assertEqual(
            self.get_repository_fields(
                'Subversion',
                fields={
                    'assembla_project_id': 'myproject',
                    'assembla_repo_name': 'myrepo',
                }
            ),
            {
                'path': 'https://subversion.assembla.com/svn/myproject/',
            })

    def test_authorize(self):
        """Testing Assembla.authorize"""
        account = self.create_hosting_account(data={})
        service = account.service

        self.assertFalse(service.is_authorized())

        service.authorize('myuser', 'abc123', None)

        self.assertIn('password', account.data)
        self.assertNotEqual(account.data['password'], 'abc123')
        self.assertTrue(service.is_authorized())

    def test_check_repository_perforce(self):
        """Testing Assembla.check_repository with Perforce"""
        try:
            account = self.create_hosting_account()
            service = account.service

            service.authorize('myuser', 'abc123', None)

            repository = Repository(hosting_account=account,
                                    tool=Tool.objects.get(name='Perforce'))
            scmtool = repository.get_scmtool()
            self.spy_on(scmtool.check_repository, call_original=False)

            service.check_repository(path='mypath',
                                     username='myusername',
                                     password='mypassword',
                                     scmtool_class=scmtool.__class__,
                                     local_site_name=None,
                                     assembla_project_id='myproject')

            self.assertTrue(scmtool.check_repository.called)
            self.assertIn('p4_host', scmtool.check_repository.last_call.kwargs)
            self.assertEqual(
                scmtool.check_repository.last_call.kwargs['p4_host'],
                'myproject')
        except ImportError:
            raise nose.SkipTest

    def test_check_repository_subversion(self):
        """Testing Assembla.check_repository with Subversion"""
        try:
            account = self.create_hosting_account()
            service = account.service

            service.authorize('myuser', 'abc123', None)

            repository = Repository(path='https://svn.example.com/',
                                    hosting_account=account,
                                    tool=Tool.objects.get(name='Subversion'))
            scmtool = repository.get_scmtool()
            self.spy_on(scmtool.check_repository, call_original=False)

            service.check_repository(path='https://svn.example.com/',
                                     username='myusername',
                                     password='mypassword',
                                     scmtool_class=scmtool.__class__,
                                     local_site_name=None)

            self.assertTrue(scmtool.check_repository.called)
            self.assertNotIn('p4_host',
                             scmtool.check_repository.last_call.kwargs)
        except ImportError:
            raise nose.SkipTest


class AssemblaFormTests(AssemblaTestCase):
    """Unit tests for reviewboard.hostingsvcs.assembla.AssemblaForm."""

    def setUp(self):
        super(AssemblaFormTests, self).setUp()

        self.account = self.create_hosting_account()

    def test_save_form_perforce(self):
        """Testing AssemblaForm with Perforce"""
        try:
            repository = self.create_repository(hosting_account=self.account,
                                                tool_name='Perforce')

            form = self.get_form(fields={'assembla_project_id': 'myproject'})
            self.spy_on(get_hostname,
                        call_fake=lambda: 'myhost.example.com')

            form.save(repository)

            self.assertIn('use_ticket_auth', repository.extra_data)
            self.assertTrue(repository.extra_data['use_ticket_auth'])
            self.assertIn('p4_host', repository.extra_data)
            self.assertIn('p4_client', repository.extra_data)
            self.assertEqual(repository.extra_data['p4_host'], 'myproject')
            self.assertEqual(repository.extra_data['p4_client'],
                             'myhost.example.com-myproject')
        except ImportError:
            raise nose.SkipTest('Perforce support is not installed')

    def test_save_form_perforce_with_portfolio(self):
        """Testing AssemblaForm with Perforce and Assembla portfolio IDs"""
        try:
            repository = self.create_repository(hosting_account=self.account,
                                                tool_name='Perforce')

            form = self.get_form(fields={
                'assembla_project_id': 'myportfolio/myproject',
            })
            self.spy_on(get_hostname,
                        call_fake=lambda: 'myhost.example.com')

            form.save(repository)

            self.assertIn('use_ticket_auth', repository.extra_data)
            self.assertTrue(repository.extra_data['use_ticket_auth'])
            self.assertIn('p4_host', repository.extra_data)
            self.assertIn('p4_client', repository.extra_data)
            self.assertEqual(repository.extra_data['p4_host'],
                             'myportfolio/myproject')
            self.assertEqual(repository.extra_data['p4_client'],
                             'myhost.example.com-myportfolio-myproject')
        except ImportError:
            raise nose.SkipTest('Perforce support is not installed')

    def test_save_form_subversion(self):
        """Testing AssemblaForm with Subversion"""
        try:
            repository = self.create_repository(
                path='https://svn.example.com/',
                hosting_account=self.account,
                tool_name='Subversion')

            form = self.get_form(fields={'assembla_project_id': 'myproject'})
            form.save(repository)

            self.assertNotIn('use_ticket_auth', repository.extra_data)
            self.assertNotIn('p4_host', repository.extra_data)
        except ImportError:
            raise nose.SkipTest



"""Views for listing and configuring integrations."""

from __future__ import unicode_literals

from djblets.integrations.views import (BaseAdminIntegrationConfigFormView,
                                        BaseAdminIntegrationListView)

from reviewboard.integrations.base import GetIntegrationManagerMixin
from reviewboard.site.urlresolvers import local_site_reverse


class AdminIntegrationConfigFormView(GetIntegrationManagerMixin,
                                     BaseAdminIntegrationConfigFormView):
    """Administration view for configuration an integration."""

    def get_success_url(self):
        """Return the URL to redirect to after saving the form.

        This will take the user back to the list of integrations.

        Returns:
            unicode:
            The URL to the integration list page.
        """
        return local_site_reverse('integration-list', request=self.request)

    def get_form_kwargs(self):
        """Return keyword arguments to pass to the form.

        This will, by default, provide ``integration`` and configuration
        ``instance`` keyword arguments to the form during initialization,
        along with the ``request`` and ``local_site`` (if any).

        Subclases can override it with additional arguments if needed.

        Returns:
            dict:
            A dictionary of keyword arguments to pass to the form.
        """
        form_kwargs = \
            super(AdminIntegrationConfigFormView, self).get_form_kwargs()
        form_kwargs['limit_to_local_site'] = \
            getattr(form_kwargs['request'], 'local_site', None)

        return form_kwargs


class AdminIntegrationListView(GetIntegrationManagerMixin,
                               BaseAdminIntegrationListView):
    """Administration view for listing integrations."""



"""Mixins for adding extra_data unit tests for API resources."""

from __future__ import unicode_literals

import json

from djblets.features.testing import override_feature_checks
from djblets.webapi.errors import INVALID_FORM_DATA
from djblets.webapi.testing.decorators import webapi_test_template

from reviewboard.webapi.base import ExtraDataAccessLevel


class ExtraDataListMixin(object):
    """Mixin for adding extra_data tests for list resources."""

    @webapi_test_template
    def test_post_with_extra_data_simple(self):
        """Testing the POST <URL> API with extra_data.key=value"""
        self.load_fixtures(self.basic_post_fixtures)

        if self.basic_post_use_admin:
            self._login_user(admin=True)

        extra_fields = {
            'extra_data.foo': 123,
            'extra_data.bar': 456,
            'extra_data.baz': '',
            'ignored': 'foo',
        }

        url, mimetype, data, objs = \
            self.setup_basic_post_test(self.user, False, None, True)
        data.update(extra_fields)

        with override_feature_checks(self.override_features):
            rsp = self.api_post(url, data, expected_mimetype=mimetype)

        self.assertEqual(rsp['stat'], 'ok')

        item_rsp = rsp[self.resource.item_result_key]

        obj = self.resource.model.objects.get(pk=item_rsp['id'])
        self.assertIn('foo', obj.extra_data)
        self.assertIn('bar', obj.extra_data)
        self.assertNotIn('baz', obj.extra_data)
        self.assertNotIn('ignored', obj.extra_data)
        self.assertEqual(obj.extra_data['foo'], extra_fields['extra_data.foo'])
        self.assertEqual(obj.extra_data['bar'], extra_fields['extra_data.bar'])

    @webapi_test_template
    def test_post_with_extra_data_json(self):
        """Testing the POST <URL> API with extra_data:json"""
        self.load_fixtures(self.basic_post_fixtures)

        if self.basic_post_use_admin:
            self._login_user(admin=True)

        url, mimetype, data, objs = \
            self.setup_basic_post_test(self.user, False, None, True)
        data['extra_data:json'] = json.dumps({
            'foo': {
                'bar': {
                    'num': 123,
                    'string': 'hi!',
                    'bool': True,
                },
            },
            'test': [1, 2, 3],
            'not_saved': None,
        })

        with override_feature_checks(self.override_features):
            rsp = self.api_post(url, data, expected_mimetype=mimetype)

        self.assertEqual(rsp['stat'], 'ok')

        item_rsp = rsp[self.resource.item_result_key]

        obj = self.resource.model.objects.get(pk=item_rsp['id'])
        self.assertIn('foo', obj.extra_data)
        self.assertIn('test', obj.extra_data)
        self.assertNotIn('not_saved', obj.extra_data)
        self.assertEqual(obj.extra_data['foo'], {
            'bar': {
                'num': 123,
                'string': 'hi!',
                'bool': True,
            },
        })
        self.assertEqual(obj.extra_data['test'], [1, 2, 3])

    @webapi_test_template
    def test_post_with_extra_data_json_patch(self):
        """Testing the POST <URL> API with extra_data:json-patch"""
        self.load_fixtures(self.basic_post_fixtures)

        if self.basic_post_use_admin:
            self._login_user(admin=True)

        url, mimetype, data, objs = \
            self.setup_basic_post_test(self.user, False, None, True)
        data['extra_data:json-patch'] = json.dumps([
            {
                'op': 'add',
                'path': '/a',
                'value': {
                    'array': [1, 2, 3],
                },
            },
            {
                'op': 'add',
                'path': '/a/b',
                'value': 'test',
            },
            {
                'op': 'add',
                'path': '/c',
                'value': None,
            },
        ])

        with override_feature_checks(self.override_features):
            rsp = self.api_post(url, data, expected_mimetype=mimetype)

        self.assertEqual(rsp['stat'], 'ok')

        item_rsp = rsp[self.resource.item_result_key]

        obj = self.resource.model.objects.get(pk=item_rsp['id'])
        self.assertIn('a', obj.extra_data)
        self.assertIn('c', obj.extra_data)
        self.assertEqual(obj.extra_data['a'], {
            'array': [1, 2, 3],
            'b': 'test',
        })
        self.assertIsNone(obj.extra_data['c'])

    @webapi_test_template
    def test_post_with_private_extra_data_simple(self):
        """Testing the POST <URL> API with private extra_data.__key=value"""
        self.load_fixtures(self.basic_post_fixtures)

        if self.basic_post_use_admin:
            self._login_user(admin=True)

        extra_fields = {
            'extra_data.__private_key': 'private_data',
        }

        url, mimetype, data, objs = self.setup_basic_post_test(
            self.user, False, None, True)
        data.update(extra_fields)

        with override_feature_checks(self.override_features):
            rsp = self.api_post(url, data, expected_mimetype=mimetype)

        self.assertEqual(rsp['stat'], 'ok')

        item_rsp = rsp[self.resource.item_result_key]

        obj = self.resource.model.objects.get(pk=item_rsp['id'])

        self.assertNotIn('__private_key', obj.extra_data)


class ExtraDataItemMixin(object):
    """Mixin for adding extra_data tests for item resources."""

    @webapi_test_template
    def test_put_with_extra_data_in_simple_form(self):
        """Testing the PUT <URL> API with extra_data.key=value"""
        self.load_fixtures(getattr(self, 'basic_put_fixtures', []))

        if self.basic_put_use_admin:
            self._login_user(admin=True)

        extra_fields = {
            'extra_data.foo': 123,
            'extra_data.bar': 456,
            'extra_data.baz': '',
            'ignored': 'foo',
        }

        url, mimetype, data, obj, objs = \
            self.setup_basic_put_test(self.user, False, None, True)
        data.update(extra_fields)

        with override_feature_checks(self.override_features):
            rsp = self.api_put(url, data, expected_mimetype=mimetype)

        self.assertEqual(rsp['stat'], 'ok')

        item_rsp = rsp[self.resource.item_result_key]

        obj = self.resource.model.objects.get(pk=item_rsp['id'])
        self.assertIn('foo', obj.extra_data)
        self.assertIn('bar', obj.extra_data)
        self.assertNotIn('baz', obj.extra_data)
        self.assertNotIn('ignored', obj.extra_data)
        self.assertEqual(obj.extra_data['foo'], extra_fields['extra_data.foo'])
        self.assertEqual(obj.extra_data['bar'], extra_fields['extra_data.bar'])

    @webapi_test_template
    def test_put_with_private_extra_data_in_simple_form(self):
        """Testing the PUT <URL> API with private extra_data.__key=value"""
        self.load_fixtures(getattr(self, 'basic_put_fixtures', []))

        if self.basic_put_use_admin:
            self._login_user(admin=True)

        extra_fields = {
            'extra_data.__private_key': 'private_data',
        }

        url, mimetype, data, obj, objs = \
            self.setup_basic_put_test(self.user, False, None, True)
        data.update(extra_fields)

        with override_feature_checks(self.override_features):
            rsp = self.api_put(url, data, expected_mimetype=mimetype)

        self.assertEqual(rsp['stat'], 'ok')

        item_rsp = rsp[self.resource.item_result_key]

        obj = self.resource.model.objects.get(pk=item_rsp['id'])

        self.assertNotIn('__private_key', obj.extra_data)

    @webapi_test_template
    def test_put_with_extra_data_json(self):
        """Testing the PUT <URL> API with extra_data:json"""
        self.load_fixtures(getattr(self, 'basic_put_fixtures', []))

        if self.basic_put_use_admin:
            self._login_user(admin=True)

        url, mimetype, data, obj, objs = \
            self.setup_basic_put_test(self.user, False, None, True)

        obj.extra_data['removed'] = 'test'

        data = {
            'extra_data:json': json.dumps({
                'foo': {
                    'bar': {
                        'num': 123,
                        'string': 'hi!',
                        'bool': True,
                    },
                },
                'test': [1, 2, 3],
                'removed': None,
            }),
        }

        with override_feature_checks(self.override_features):
            rsp = self.api_put(url, data, expected_mimetype=mimetype)

        self.assertEqual(rsp['stat'], 'ok')

        item_rsp = rsp[self.resource.item_result_key]

        obj = self.resource.model.objects.get(pk=item_rsp['id'])
        self.assertIn('foo', obj.extra_data)
        self.assertIn('test', obj.extra_data)
        self.assertNotIn('removed', obj.extra_data)
        self.assertEqual(obj.extra_data['foo'], {
            'bar': {
                'num': 123,
                'string': 'hi!',
                'bool': True,
            },
        })
        self.assertEqual(obj.extra_data['test'], [1, 2, 3])

    @webapi_test_template
    def test_put_with_extra_data_json_with_private_keys(self):
        """Testing the PUT <URL> API with extra_data:json with private keys"""
        self.load_fixtures(getattr(self, 'basic_put_fixtures', []))

        if self.basic_put_use_admin:
            self._login_user(admin=True)

        url, mimetype, data, obj, objs = \
            self.setup_basic_put_test(self.user, False, None, True)

        data['extra_data:json'] = json.dumps({
            'foo': {
                '__bar': {
                    'num': 123,
                    'string': 'hi!',
                    'bool': True,
                },
                'baz': 456,
            },
        })

        with override_feature_checks(self.override_features):
            rsp = self.api_put(url, data, expected_mimetype=mimetype)

        self.assertEqual(rsp['stat'], 'ok')

        item_rsp = rsp[self.resource.item_result_key]

        obj = self.resource.model.objects.get(pk=item_rsp['id'])
        self.assertIn('foo', obj.extra_data)
        self.assertEqual(obj.extra_data['foo'], {
            'baz': 456,
        })

    @webapi_test_template
    def test_put_with_extra_data_json_with_access_levels(self):
        """Testing the PUT <URL> API with extra_data:json with access levels"""
        def _access_cb(path):
            if path == ('parent', 'private'):
                return ExtraDataAccessLevel.ACCESS_STATE_PRIVATE
            elif path == ('public-readonly',):
                return ExtraDataAccessLevel.ACCESS_STATE_PUBLIC_READONLY
            else:
                return ExtraDataAccessLevel.ACCESS_STATE_PUBLIC

        self.load_fixtures(getattr(self, 'basic_put_fixtures', []))

        if self.basic_put_use_admin:
            self._login_user(admin=True)

        url, mimetype, data, obj, objs = \
            self.setup_basic_put_test(self.user, False, None, True)

        obj.extra_data = {
            'public-readonly': 'orig',
        }
        obj.save()

        try:
            self.resource.extra_data_access_callbacks.register(_access_cb)

            data = {
                'extra_data:json': json.dumps({
                    'parent': {
                        'private': 1,
                    },
                    'public-readonly': 2,
                    'test': 3,
                })
            }

            with override_feature_checks(self.override_features):
                rsp = self.api_put(url, data, expected_mimetype=mimetype)
        finally:
            self.resource.extra_data_access_callbacks.unregister(_access_cb)

        self.assertEqual(rsp['stat'], 'ok')

        item_rsp = rsp[self.resource.item_result_key]

        obj = self.resource.model.objects.get(pk=item_rsp['id'])
        self.assertEqual(obj.extra_data, {
            'parent': {},
            'public-readonly': 'orig',
            'test': 3,
        })

    @webapi_test_template
    def test_put_with_extra_data_json_with_override_root(self):
        """Testing the PUT <URL> API with extra_data:json with attempting to
        override root of extra_data
        """
        self.load_fixtures(getattr(self, 'basic_put_fixtures', []))

        if self.basic_put_use_admin:
            self._login_user(admin=True)

        url, mimetype, data, obj, objs = \
            self.setup_basic_put_test(self.user, False, None, True)

        obj.extra_data = {
            'a': 1,
            'b': 2,
        }
        obj.save()

        data['extra_data:json'] = json.dumps([1, 2, 3])

        with override_feature_checks(self.override_features):
            rsp = self.api_put(url, data, expected_status=400)

        self.assertEqual(rsp['stat'], 'fail')
        self.assertEqual(rsp['err'], {
            'code': INVALID_FORM_DATA.code,
            'msg': 'One or more fields had errors',
        })
        self.assertEqual(rsp['fields'], {
            'extra_data': [
                'extra_data:json cannot replace extra_data with a '
                'non-dictionary type',
            ],
        })

        obj = self.resource.model.objects.get(pk=obj.pk)
        self.assertEqual(obj.extra_data, {
            'a': 1,
            'b': 2,
        })

    @webapi_test_template
    def test_put_with_extra_data_json_patch(self):
        """Testing the PUT <URL> API with extra_data:json-patch"""
        self.load_fixtures(getattr(self, 'basic_put_fixtures', []))

        if self.basic_put_use_admin:
            self._login_user(admin=True)

        url, mimetype, data, obj, objs = \
            self.setup_basic_put_test(self.user, False, None, True)

        obj.extra_data = {
            'a': 1,
            'b': {
                'c': 2,
            },
        }
        obj.save()

        data = {
            'extra_data:json-patch': json.dumps([
                {
                    'op': 'add',
                    'path': '/b/d',
                    'value': 3,
                },
                {
                    'op': 'remove',
                    'path': '/a',
                },
                {
                    'op': 'copy',
                    'from': '/b/c',
                    'path': '/e',
                },
            ]),
        }

        with override_feature_checks(self.override_features):
            rsp = self.api_put(url, data, expected_mimetype=mimetype)

        self.assertEqual(rsp['stat'], 'ok')

        item_rsp = rsp[self.resource.item_result_key]

        obj = self.resource.model.objects.get(pk=item_rsp['id'])
        self.assertEqual(obj.extra_data, {
            'b': {
                'c': 2,
                'd': 3,
            },
            'e': 2,
        })

    @webapi_test_template
    def test_put_with_extra_data_json_patch_with_private_keys(self):
        """Testing the PUT <URL> API with extra_data:json-patch with private
        keys
        """
        self.load_fixtures(getattr(self, 'basic_put_fixtures', []))

        if self.basic_put_use_admin:
            self._login_user(admin=True)

        url, mimetype, data, obj, objs = \
            self.setup_basic_put_test(self.user, False, None, True)

        obj.extra_data = {
            'a': 1,
            'b': {
                '__c': 2,
            },
        }
        obj.save()

        data['extra_data:json-patch'] = json.dumps([
            {
                'op': 'remove',
                'path': '/b/__c',
            },
        ])

        with override_feature_checks(self.override_features):
            rsp = self.api_put(url, data, expected_status=400)

        self.assertEqual(rsp['stat'], 'fail')
        self.assertEqual(rsp['err'], {
            'code': INVALID_FORM_DATA.code,
            'msg': 'One or more fields had errors',
        })
        self.assertEqual(rsp['fields'], {
            'extra_data': [
                'Failed to patch JSON data: Cannot write to path "/b/__c" '
                'for patch entry 0',
            ],
        })

        obj = self.resource.model.objects.get(pk=obj.pk)
        self.assertEqual(obj.extra_data, {
            'a': 1,
            'b': {
                '__c': 2,
            },
        })

    @webapi_test_template
    def test_put_with_extra_data_json_patch_with_private_access_level(self):
        """Testing the PUT <URL> API with extra_data:json-patch with private
        access level
        """
        def _access_cb(path):
            if path == ('a',):
                return ExtraDataAccessLevel.ACCESS_STATE_PRIVATE

            return ExtraDataAccessLevel.ACCESS_STATE_PUBLIC

        self.load_fixtures(getattr(self, 'basic_put_fixtures', []))

        if self.basic_put_use_admin:
            self._login_user(admin=True)

        url, mimetype, data, obj, objs = \
            self.setup_basic_put_test(self.user, False, None, True)

        obj.extra_data = {
            'a': 1,
        }
        obj.save()

        try:
            self.resource.extra_data_access_callbacks.register(_access_cb)

            data['extra_data:json-patch'] = json.dumps([
                {
                    'op': 'add',
                    'path': '/b',
                    'value': 1,
                },
                {
                    'op': 'remove',
                    'path': '/a',
                },
            ])

            with override_feature_checks(self.override_features):
                rsp = self.api_put(url, data, expected_status=400)
        finally:
            self.resource.extra_data_access_callbacks.unregister(_access_cb)

        self.assertEqual(rsp['stat'], 'fail')
        self.assertEqual(rsp['err'], {
            'code': INVALID_FORM_DATA.code,
            'msg': 'One or more fields had errors',
        })
        self.assertEqual(rsp['fields'], {
            'extra_data': [
                'Failed to patch JSON data: Cannot write to path "/a" '
                'for patch entry 1',
            ],
        })

        obj = self.resource.model.objects.get(pk=obj.pk)
        self.assertEqual(obj.extra_data, {
            'a': 1,
        })

    @webapi_test_template
    def test_put_with_extra_data_json_patch_with_read_only_access_level(self):
        """Testing the PUT <URL> API with extra_data:json-patch with read-only
        access level
        """
        def _access_cb(path):
            if path == ('a',):
                return ExtraDataAccessLevel.ACCESS_STATE_PUBLIC_READONLY

            return ExtraDataAccessLevel.ACCESS_STATE_PUBLIC

        self.load_fixtures(getattr(self, 'basic_put_fixtures', []))

        if self.basic_put_use_admin:
            self._login_user(admin=True)

        url, mimetype, data, obj, objs = \
            self.setup_basic_put_test(self.user, False, None, True)

        obj.extra_data = {
            'a': 1,
        }
        obj.save()

        try:
            self.resource.extra_data_access_callbacks.register(_access_cb)

            data['extra_data:json-patch'] = json.dumps([
                {
                    'op': 'test',
                    'path': '/a',
                    'value': 1,
                },
                {
                    'op': 'copy',
                    'from': '/a',
                    'path': '/b',
                },
                {
                    'op': 'remove',
                    'path': '/a',
                },
            ])

            with override_feature_checks(self.override_features):
                rsp = self.api_put(url, data, expected_status=400)
        finally:
            self.resource.extra_data_access_callbacks.unregister(_access_cb)

        self.assertEqual(rsp['stat'], 'fail')
        self.assertEqual(rsp['err'], {
            'code': INVALID_FORM_DATA.code,
            'msg': 'One or more fields had errors',
        })
        self.assertEqual(rsp['fields'], {
            'extra_data': [
                'Failed to patch JSON data: Cannot write to path "/a" '
                'for patch entry 2',
            ],
        })

        obj = self.resource.model.objects.get(pk=obj.pk)
        self.assertEqual(obj.extra_data, {
            'a': 1,
        })

    @webapi_test_template
    def test_put_with_extra_data_json_patch_with_override_root(self):
        """Testing the PUT <URL> API with extra_data:json-patch with
        attempting to override root of extra_data
        """
        self.load_fixtures(getattr(self, 'basic_put_fixtures', []))

        if self.basic_put_use_admin:
            self._login_user(admin=True)

        url, mimetype, data, obj, objs = \
            self.setup_basic_put_test(self.user, False, None, True)

        obj.extra_data = {
            'a': 1,
            'b': 2,
        }
        obj.save()

        data['extra_data:json-patch'] = json.dumps([
            {
                'op': 'replace',
                'path': '',
                'value': {
                    'new': 'values',
                },
            },
        ])

        with override_feature_checks(self.override_features):
            rsp = self.api_put(url, data, expected_status=400)

        self.assertEqual(rsp['stat'], 'fail')
        self.assertEqual(rsp['err'], {
            'code': INVALID_FORM_DATA.code,
            'msg': 'One or more fields had errors',
        })
        self.assertEqual(rsp['fields'], {
            'extra_data': [
                'Failed to patch JSON data: Cannot write to path "" '
                'for patch entry 0',
            ],
        })

        obj = self.resource.model.objects.get(pk=obj.pk)
        self.assertEqual(obj.extra_data, {
            'a': 1,
            'b': 2,
        })

    @webapi_test_template
    def test_get_with_private_extra_data_in_key_form(self):
        """Testing the GET <URL> API with private extra_data in __key form"""
        self.load_fixtures(getattr(self, 'basic_put_fixtures', []))

        if self.basic_get_use_admin:
            self._login_user(admin=True)

        extra_fields = {
            '__private_key': 'private_data',
            'public_key': {
                '__another_private_key': 'foo',
                'another_public_key': 'bar',
            },
        }

        url, mimetype, item = self.setup_basic_get_test(self.user, False, None)

        obj = self.resource.model.objects.get(pk=item.id)
        obj.extra_data = extra_fields
        obj.save(update_fields=['extra_data'])

        with override_feature_checks(self.override_features):
            rsp = self.api_get(url,
                               expected_mimetype=mimetype,
                               expected_json=self.basic_get_returns_json)

        item_rsp = rsp[self.resource.item_result_key]

        self.assertNotIn('__private_key', item_rsp['extra_data'])
        self.assertNotIn('__another_private_key',
                         item_rsp['extra_data']['public_key'])
        self.assertIn('another_public_key',
                      item_rsp['extra_data']['public_key'])



from __future__ import unicode_literals

from django.utils import six

from reviewboard.hostingsvcs.service import (get_hosting_services,
                                             get_hosting_service)
from reviewboard.webapi.resources import resources
from reviewboard.webapi.tests.base import BaseWebAPITestCase
from reviewboard.webapi.tests.mimetypes import (hosting_service_item_mimetype,
                                                hosting_service_list_mimetype)
from reviewboard.webapi.tests.mixins import BasicTestsMetaclass
from reviewboard.webapi.tests.urls import (get_hosting_service_item_url,
                                           get_hosting_service_list_url)


def _compare_item(self, item_rsp, hosting_service):
    self.assertEqual(item_rsp['id'], hosting_service.hosting_service_id)
    self.assertEqual(item_rsp['name'], hosting_service.name)
    self.assertEqual(item_rsp['needs_authorization'],
                     hosting_service.needs_authorization)
    self.assertEqual(item_rsp['supports_bug_trackers'],
                     hosting_service.supports_bug_trackers)
    self.assertEqual(item_rsp['supports_repositories'],
                     hosting_service.supports_repositories)
    self.assertEqual(item_rsp['supports_two_factor_auth'],
                     hosting_service.supports_two_factor_auth)
    self.assertEqual(item_rsp['supported_scmtools'],
                     hosting_service.supported_scmtools)

    # Compute the base URL for links.
    url_base = 'http://testserver/'

    if '/s/local-site-1/' in item_rsp['links']['self']['href']:
        url_base += 's/local-site-1/'

    url_base += 'api/'

    # Check the links.
    accounts_url = url_base + ('hosting-service-accounts/?service=%s'
                               % hosting_service.hosting_service_id)
    self.assertIn('accounts', item_rsp['links'])
    self.assertEqual(item_rsp['links']['accounts']['href'], accounts_url)

    accounts_url = url_base + ('repositories/?hosting-service=%s'
                               % hosting_service.hosting_service_id)
    self.assertIn('repositories', item_rsp['links'])
    self.assertEqual(item_rsp['links']['repositories']['href'], accounts_url)


@six.add_metaclass(BasicTestsMetaclass)
class ResourceListTests(BaseWebAPITestCase):
    """Testing the HostingServiceResource list APIs."""
    fixtures = ['test_users']
    sample_api_url = 'hosting-services/'
    resource = resources.hosting_service

    compare_item = _compare_item

    def setup_http_not_allowed_list_test(self, user):
        return get_hosting_service_list_url()

    #
    # HTTP GET tests
    #

    def setup_basic_get_test(self, user, with_local_site, local_site_name,
                             populate_items):
        return (get_hosting_service_list_url(local_site_name),
                hosting_service_list_mimetype,
                get_hosting_services())


@six.add_metaclass(BasicTestsMetaclass)
class ResourceItemTests(BaseWebAPITestCase):
    """Testing the HostingServiceResource item APIs."""
    fixtures = ['test_users']
    sample_api_url = 'hosting-services/<id>/'
    resource = resources.hosting_service

    compare_item = _compare_item

    def setup_http_not_allowed_item_test(self, user):
        hosting_service = get_hosting_service('github')

        return get_hosting_service_item_url(hosting_service)

    #
    # HTTP GET tests
    #

    def setup_basic_get_test(self, user, with_local_site, local_site_name):
        hosting_service = get_hosting_service('github')

        return (get_hosting_service_item_url(hosting_service, local_site_name),
                hosting_service_item_mimetype,
                hosting_service)



from __future__ import unicode_literals

from django.contrib.auth.decorators import login_required
from django.contrib.auth.models import User
from django.http import Http404
from django.shortcuts import get_object_or_404, render
from django.utils.translation import ugettext_lazy as _

from reviewboard.accounts.decorators import (check_login_required,
                                             valid_prefs_required)
from reviewboard.datagrids.grids import (DashboardDataGrid,
                                         GroupDataGrid,
                                         ReviewRequestDataGrid,
                                         UsersDataGrid,
                                         UserPageReviewsDataGrid,
                                         UserPageReviewRequestDataGrid)
from reviewboard.reviews.models import Group, ReviewRequest
from reviewboard.site.decorators import check_local_site_access
from reviewboard.site.urlresolvers import local_site_reverse


def _is_datagrid_gridonly(request):
    """Return whether or not the current request is for an embedded datagrid.

    This method allows us to disable consent checks in
    :py:func:`~reviewboard.accounts.decorators.valid_prefs_required` when a
    datagrid is requesting updated data so that we do not return a redirect
    and embed the result of that redirect in the datagrid instead.

    Args:
        request (django.http.HttpRequest):
            The HTTP request from the client.

    Returns:
        bool:
        Whether or not this request is for an embedded datagrid.
    """
    return 'gridonly' in request.GET


@check_login_required
@check_local_site_access
@valid_prefs_required(disable_consent_checks=_is_datagrid_gridonly)
def all_review_requests(request,
                        local_site=None,
                        template_name='datagrids/datagrid.html'):
    """Display a list of all review requests."""
    datagrid = ReviewRequestDataGrid(
        request,
        ReviewRequest.objects.public(user=request.user,
                                     status=None,
                                     local_site=local_site,
                                     with_counts=True,
                                     show_inactive=True),
        _("All Review Requests"),
        local_site=local_site)
    return datagrid.render_to_response(template_name)


@login_required
@check_local_site_access
@valid_prefs_required(disable_consent_checks=_is_datagrid_gridonly)
def dashboard(request,
              template_name='datagrids/dashboard.html',
              local_site=None):
    """Display the dashboard.

    This shows review requests organized by a variety of lists, depending on
    the 'view' parameter.

    Valid 'view' parameters are:

        * 'outgoing'
        * 'to-me'
        * 'to-group'
        * 'starred'
        * 'incoming'
        * 'mine'
    """
    grid = DashboardDataGrid(request, local_site=local_site)
    return grid.render_to_response(template_name)


@check_login_required
@check_local_site_access
@valid_prefs_required(disable_consent_checks=_is_datagrid_gridonly)
def group(request,
          name,
          template_name='datagrids/datagrid.html',
          local_site=None):
    """Display a list of review requests belonging to a particular group."""
    # Make sure the group exists
    group = get_object_or_404(Group, name=name, local_site=local_site)

    if not group.is_accessible_by(request.user):
        return render(request, 'datagrids/group_permission_denied.html',
                      status=403)

    datagrid = ReviewRequestDataGrid(
        request,
        ReviewRequest.objects.to_group(name,
                                       local_site,
                                       user=request.user,
                                       status=None,
                                       with_counts=True),
        _('Review requests for %s') % group.display_name,
        local_site=local_site)

    return datagrid.render_to_response(template_name)


@check_login_required
@check_local_site_access
@valid_prefs_required(disable_consent_checks=_is_datagrid_gridonly)
def group_list(request,
               local_site=None,
               template_name='datagrids/datagrid.html'):
    """Display a list of all review groups."""
    grid = GroupDataGrid(request, local_site=local_site)
    return grid.render_to_response(template_name)


@check_login_required
@check_local_site_access
@valid_prefs_required(disable_consent_checks=_is_datagrid_gridonly)
def group_members(request,
                  name,
                  template_name='datagrids/datagrid.html',
                  local_site=None):
    """Display a list of users registered for a particular group."""
    # Make sure the group exists
    group = get_object_or_404(Group,
                              name=name,
                              local_site=local_site)

    if not group.is_accessible_by(request.user):
        return render(request, 'datagrids/group_permission_denied.html',
                      status=403)

    datagrid = UsersDataGrid(request,
                             group.users.filter(is_active=True),
                             _("Members of group %s") % name)

    return datagrid.render_to_response(template_name)


@check_login_required
@check_local_site_access
@valid_prefs_required(disable_consent_checks=_is_datagrid_gridonly)
def submitter(request,
              username,
              grid=None,
              template_name='datagrids/datagrid.html',
              local_site=None):
    """Display a user's profile, showing their review requests and reviews.

    The 'grid' parameter determines which is displayed, and can take on the
    following values:

        * 'reviews'
        * 'review-requests'
    """
    # Make sure the user exists
    if local_site:
        try:
            user = local_site.users.get(username=username)
        except User.DoesNotExist:
            raise Http404
    else:
        user = get_object_or_404(User, username=username)

    if grid is None or grid == 'review-requests':
        datagrid_cls = UserPageReviewRequestDataGrid
    elif grid == 'reviews':
        datagrid_cls = UserPageReviewsDataGrid
    else:
        raise Http404

    datagrid = datagrid_cls(request, user, local_site=local_site)
    datagrid.tabs = [
        (UserPageReviewRequestDataGrid.tab_title,
         local_site_reverse('user', local_site=local_site,
                            args=[username])),
        (UserPageReviewsDataGrid.tab_title,
         local_site_reverse('user-grid', local_site=local_site,
                            args=[username, 'reviews']))
    ]

    return datagrid.render_to_response(template_name)


@check_login_required
@check_local_site_access
@valid_prefs_required(disable_consent_checks=_is_datagrid_gridonly)
def users_list(request,
               local_site=None,
               template_name='datagrids/datagrid.html'):
    """Display a list of all users."""
    grid = UsersDataGrid(request, local_site=local_site)
    return grid.render_to_response(template_name)



"""Definitions for the StatusUpdate model."""

from __future__ import unicode_literals

import datetime

from django.contrib.auth.models import User
from django.db import models
from django.utils import timezone
from django.utils.translation import ugettext_lazy as _
from djblets.db.fields import JSONField

from reviewboard.changedescs.models import ChangeDescription
from reviewboard.reviews.models.review import Review
from reviewboard.reviews.models.review_request import ReviewRequest


class StatusUpdate(models.Model):
    """A status update from a third-party service or extension.

    This status model allows external services (such as continuous integration
    services, Review Bot, etc.) to provide an update on their status. An
    example of this would be a CI tool which does experimental builds of
    changes. While the build is running, that tool would set its status to
    pending, and when it was done, would set it to one of the done states,
    and potentially associate it with a review containing issues.
    """

    #: The pending state.
    PENDING = 'P'

    #: The completed successfully state.
    DONE_SUCCESS = 'S'

    #: The completed with reported failures state.
    DONE_FAILURE = 'F'

    #: The error state.
    ERROR = 'E'

    #: Timeout state.
    TIMEOUT = 'T'

    STATUSES = (
        (PENDING, _('Pending')),
        (DONE_SUCCESS, _('Done (Success)')),
        (DONE_FAILURE, _('Done (Failure)')),
        (ERROR, _('Error')),
        (TIMEOUT, _('Timed Out')),
    )

    #: An identifier for the service posting this status update.
    #:
    #: This ID is self-assigned, and just needs to be unique to that service.
    #: Possible values can be an extension ID, webhook URL, or a script name.
    service_id = models.CharField(_('Service ID'), max_length=255)

    #: The user who created this status update.
    user = models.ForeignKey(
        User,
        related_name='status_updates',
        verbose_name=_('User'),
        blank=True,
        null=True)

    #: The timestamp of the status update.
    timestamp = models.DateTimeField(_('Timestamp'), auto_now=True)

    #: A user-visible short summary of the status update.
    #:
    #: This is typically the name of the integration or tool that was run.
    summary = models.CharField(_('Summary'), max_length=255)

    #: A user-visible description on the status update.
    #:
    #: This is shown in the UI adjacent to the summary. Typical results might
    #: be things like "running." or "failed.". This should include punctuation.
    description = models.CharField(_('Description'), max_length=255,
                                   blank=True)

    #: An optional link.
    #:
    #: This is used in case the tool has some external page, such as a build
    #: results page on a CI system.
    url = models.URLField(_('Link URL'), max_length=255, blank=True)

    #: Text for the link. If ``url`` is empty, this will not be used.
    url_text = models.CharField(_('Link text'), max_length=64, blank=True)

    #: The current state of this status update.
    #:
    #: This should be set to :py:attr:`PENDING` while the service is
    #: processing the update, and then to either :py:attr:`DONE_SUCCESS` or
    #: :py:attr:`DONE_FAILURE` once complete. If the service encountered some
    #: error which prevented completion, this should be set to
    #: :py:attr:`ERROR`.
    state = models.CharField(_('State'), max_length=1, choices=STATUSES)

    #: The review request that this status update is for.
    review_request = models.ForeignKey(
        ReviewRequest,
        related_name='status_updates',
        verbose_name=_('Review Request'))

    #: The change to the review request that this status update is for.
    #:
    #: If this is ``None``, this status update refers to the review request as
    #: a whole (for example, the initial diff that was posted).
    change_description = models.ForeignKey(
        ChangeDescription,
        related_name='status_updates',
        verbose_name=_('Change Description'),
        null=True,
        blank=True)

    #: An optional review created for this status update.
    #:
    #: This allows the third-party service to create comments and open issues.
    review = models.OneToOneField(
        Review,
        related_name='status_update',
        verbose_name=_('Review'),
        null=True,
        blank=True)

    #: Any extra data that the service wants to store for this status update.
    extra_data = JSONField(null=True)

    #: An (optional) timeout, in seconds. If this is non-None and the state has
    #: been ``PENDING`` for longer than this period (computed from the
    #: :py:attr:`timestamp` field), :py:attr:`effective_state` will be
    #: ``TIMEOUT``.
    timeout = models.IntegerField(null=True, blank=True)

    @staticmethod
    def state_to_string(state):
        """Return a string representation of a status update state.

        Args:
            state (unicode):
                A single-character string representing the state.

        Returns:
            unicode:
            A longer string representation of the state suitable for use in
            the API.
        """
        if state == StatusUpdate.PENDING:
            return 'pending'
        elif state == StatusUpdate.DONE_SUCCESS:
            return 'done-success'
        elif state == StatusUpdate.DONE_FAILURE:
            return 'done-failure'
        elif state == StatusUpdate.ERROR:
            return 'error'
        elif state == StatusUpdate.TIMEOUT:
            return 'timed-out'
        else:
            raise ValueError('Invalid state "%s"' % state)

    @staticmethod
    def string_to_state(state):
        """Return a status update state from an API string.

        Args:
            state (unicode):
                A string from the API representing the state.

        Returns:
            unicode:
            A single-character string representing the state, suitable for
            storage in the ``state`` field.
        """
        if state == 'pending':
            return StatusUpdate.PENDING
        elif state == 'done-success':
            return StatusUpdate.DONE_SUCCESS
        elif state == 'done-failure':
            return StatusUpdate.DONE_FAILURE
        elif state == 'error':
            return StatusUpdate.ERROR
        elif state == 'timed-out':
            return StatusUpdate.TIMEOUT
        else:
            raise ValueError('Invalid state string "%s"' % state)

    def is_mutable_by(self, user):
        """Return whether the user can modify this status update.

        Args:
            user (django.contrib.auth.models.User):
                The user to check.

        Returns:
            bool:
            True if the user can modify this status update.
        """
        return (self.user == user or
                user.has_perm('reviews.can_edit_status',
                              self.review_request.local_site))

    @property
    def effective_state(self):
        """The state of the status update, taking into account timeouts."""
        if self.state == self.PENDING and self.timeout is not None:
            timeout = self.timestamp + datetime.timedelta(seconds=self.timeout)

            if timezone.now() > timeout:
                return self.TIMEOUT

        return self.state

    class Meta:
        app_label = 'reviews'
        db_table = 'reviews_statusupdate'
        ordering = ['timestamp']
        get_latest_by = 'timestamp'
        verbose_name = _('Status Update')
        verbose_name_plural = _('Status Updates')



from __future__ import unicode_literals

import os

import nose
from django.utils import six
from djblets.util.filesystem import is_exe_in_path

from reviewboard.scmtools.errors import (FileNotFoundError,
                                         InvalidRevisionFormatError,
                                         RepositoryNotFoundError, SCMError)
from reviewboard.scmtools.models import Repository, Tool
from reviewboard.scmtools.tests.testcases import SCMTestCase


class BZRTests(SCMTestCase):
    """Unit tests for bzr."""

    fixtures = ['test_scmtools']

    def setUp(self):
        super(BZRTests, self).setUp()

        if not is_exe_in_path('bzr'):
            raise nose.SkipTest()

        self.bzr_repo_path = os.path.join(os.path.dirname(__file__),
                                          '..', 'testdata', 'bzr_repo')
        self.bzr_ssh_path = ('bzr+ssh://localhost/%s'
                             % self.bzr_repo_path.replace('\\', '/'))
        self.bzr_sftp_path = ('sftp://localhost/%s'
                              % self.bzr_repo_path.replace('\\', '/'))
        self.repository = Repository(name='Bazaar',
                                     path='file://' + self.bzr_repo_path,
                                     tool=Tool.objects.get(name='Bazaar'))

        self.tool = self.repository.get_scmtool()

    def test_check_repository(self):
        """Testing BZRTool.check_repository"""
        self.tool.check_repository(self.repository.path)

    def test_check_repository_with_not_found(self):
        """Testing BZRTool.check_repository with repository not found"""
        with self.assertRaises(RepositoryNotFoundError):
            self.tool.check_repository('file:///dummy')

    def test_ssh(self):
        """Testing BZRTool with a SSH-backed repository"""
        self._test_ssh(self.bzr_ssh_path, 'README')

    def test_ssh_with_site(self):
        """Testing BZRTool with a SSH-backed repository with a LocalSite"""
        self._test_ssh_with_site(self.bzr_ssh_path, 'README')

    def test_sftp(self):
        """Testing BZRTool with a SFTP-backed repository"""
        try:
            self._test_ssh(self.bzr_sftp_path, 'README')
        except SCMError as e:
            err = six.text_type(e)

            if 'Installed bzr and paramiko modules are incompatible' in err:
                raise nose.SkipTest(err)

            raise

    def test_get_file(self):
        """Testing BZRTool.get_file"""
        content = self.tool.get_file('README', '2011-02-02 10:53:04 +0000')
        self.assertEqual(content, b'This is a test.\n')
        self.assertIsInstance(content, bytes)

    def test_get_file_with_timezone_offset(self):
        """Testing BZRTool.get_file with timezone offset"""
        content = self.tool.get_file('README', '2011-02-02 02:53:04 -0800')
        self.assertEqual(content, b'This is a test.\n')
        self.assertIsInstance(content, bytes)

    def test_get_file_with_non_utc_server_timezone(self):
        """Testing BZRTool.get_file with settings.TIME_ZONE != UTC"""
        old_timezone = os.environ[b'TZ']
        os.environ[b'TZ'] = b'US/Pacific'

        try:
            content = self.tool.get_file('README', '2011-02-02 02:53:04 -0800')
        finally:
            os.environ[b'TZ'] = old_timezone

        self.assertEqual(content, b'This is a test.\n')
        self.assertIsInstance(content, bytes)

    def test_get_file_with_revision_id(self):
        """Testing BZRTool.get_file with revision ID"""
        content = self.tool.get_file(
            'README',
            'revid:chipx86@chipx86.com-20110202105304-8lkgyb18aqr11b21')
        self.assertEqual(content, b'This is a test.\n')
        self.assertIsInstance(content, bytes)

    def test_get_file_with_unknown_file(self):
        """Testing BZRTool.get_file with unknown file"""
        with self.assertRaises(FileNotFoundError):
            self.tool.get_file('NOT_REAL', '2011-02-02 02:53:04 -0800')

    def test_get_file_with_unknown_revision(self):
        """Testing BZRTool.get_file with unknown revision"""
        with self.assertRaises(FileNotFoundError):
            self.tool.get_file('README', '9999-02-02 02:53:04 -0800')

    def test_get_file_with_invalid_revision(self):
        """Testing BZRTool.get_file with invalid revision"""
        with self.assertRaises(InvalidRevisionFormatError):
            self.tool.get_file('README', '\o/')

    def test_file_exists(self):
        """Testing BZRTool.files_exists"""
        self.assertTrue(self.tool.file_exists(
            'README',
            '2011-02-02 10:53:04 +0000'))

        self.assertFalse(self.tool.file_exists(
            'NOT_REAL',
            '2011-02-02 10:53:04 +0000'))
        self.assertFalse(self.tool.file_exists(
            'README',
            '9999-02-02 10:53:04 +0000'))

    def test_file_exists_with_timezone_offset(self):
        """Testing BZRTool.files_exists with timezone offset"""
        self.assertTrue(self.tool.file_exists(
            'README',
            '2011-02-02 02:53:04 -0800'))

    def test_file_exists_with_non_utc_server_timezone(self):
        """Testing BZRTool.files_exists with settings.TIME_ZONE != UTC"""
        old_timezone = os.environ[b'TZ']
        os.environ[b'TZ'] = b'US/Pacific'

        try:
            self.assertTrue(self.tool.file_exists(
                'README',
                '2011-02-02 02:53:04 -0800'))
        finally:
            os.environ[b'TZ'] = old_timezone

    def test_file_exists_with_revision_id(self):
        """Testing BZRTool.files_exists with revision ID"""
        self.assertTrue(self.tool.file_exists(
            'README',
            'revid:chipx86@chipx86.com-20110202105304-8lkgyb18aqr11b21'))

    def test_file_exists_with_invalid_revision(self):
        """Testing BZRTool.files_exists with invalid revision"""
        with self.assertRaises(InvalidRevisionFormatError):
            self.tool.file_exists('README', '\o/')



from __future__ import unicode_literals

import os
from errno import ECONNREFUSED
from socket import error as SocketError
from tempfile import mkdtemp

import nose

from reviewboard.scmtools.core import HEAD
from reviewboard.scmtools.errors import SCMError, AuthenticationError
from reviewboard.scmtools.models import Repository
from reviewboard.site.models import LocalSite
from reviewboard.ssh.client import SSHClient
from reviewboard.ssh.tests import SSHTestCase


class SCMTestCase(SSHTestCase):
    """Base class for test suites for SCMTools."""

    ssh_client = None
    _can_test_ssh = None

    def setUp(self):
        super(SCMTestCase, self).setUp()
        self.tool = None

    def _check_can_test_ssh(self):
        """Check whether SSH-based tests can be run.

        This will check if the user's SSH keys is authorized by the local
        machine, for authentication. If so, SSH-based tests can be attempted.

        If SSH-based tests cannot be run, the current test will be flagged
        as skipped.
        """
        if SCMTestCase._can_test_ssh is None:
            SCMTestCase.ssh_client = SSHClient()
            key = self.ssh_client.get_user_key()
            SCMTestCase._can_test_ssh = \
                key is not None and self.ssh_client.is_key_authorized(key)

        if not SCMTestCase._can_test_ssh:
            raise nose.SkipTest(
                "Cannot perform SSH access tests. The local user's SSH "
                "public key must be in the %s file and SSH must be enabled."
                % os.path.join(self.ssh_client.storage.get_ssh_dir(),
                               'authorized_keys'))

    def _test_ssh(self, repo_path, filename=None):
        """Helper for testing an SSH connection to a local repository.

        This will attempt to SSH into the local machine and connect to the
        given repository, checking it for validity and optionally fetching
        a file.

        If this is unable to connect to the local machine, the test will be
        flagged as skipped.

        Args:
            repo_path (unicode):
                The repository path to check.

            filename (unicode, optional):
                The optional file in the repository to fetch.
        """
        self._check_can_test_ssh()

        repo = Repository(name='SSH Test', path=repo_path,
                          tool=self.repository.tool)
        tool = repo.get_scmtool()

        try:
            tool.check_repository(repo_path)
        except SocketError as e:
            if e.errno == ECONNREFUSED:
                # This box likely isn't set up for this test.
                SCMTestCase._can_test_ssh = False

                raise nose.SkipTest(
                    'Cannot perform SSH access tests. No local SSH service is '
                    'running.')
            else:
                raise

        if filename:
            self.assertNotEqual(tool.get_file(filename, HEAD), None)

    def _test_ssh_with_site(self, repo_path, filename=None):
        """Helper for testing an SSH connection and using a Local Site.

        This will attempt to SSH into the local machine and connect to the
        given repository, using an SSH key and repository based on a Local
        Site. It will check the repository for validity and optionally fetch
        a file.

        If this is unable to connect to the local machine, the test will be
        flagged as skipped.

        Args:
            repo_path (unicode):
                The repository path to check.

            filename (unicode, optional):
                The optional file in the repository to fetch.
        """
        self._check_can_test_ssh()

        # Get the user's .ssh key, for use in the tests
        user_key = self.ssh_client.get_user_key()
        self.assertNotEqual(user_key, None)

        # Switch to a new SSH directory.
        self.tempdir = mkdtemp(prefix='rb-tests-home-')
        sshdir = os.path.join(self.tempdir, '.ssh')
        self._set_home(self.tempdir)

        self.assertEqual(sshdir, self.ssh_client.storage.get_ssh_dir())
        self.assertFalse(os.path.exists(os.path.join(sshdir, 'id_rsa')))
        self.assertFalse(os.path.exists(os.path.join(sshdir, 'id_dsa')))
        self.assertEqual(self.ssh_client.get_user_key(), None)

        tool_class = self.repository.tool

        # Make sure we aren't using the old SSH key. We want auth errors.
        repo = Repository(name='SSH Test', path=repo_path, tool=tool_class)
        tool = repo.get_scmtool()
        self.assertRaises(AuthenticationError,
                          lambda: tool.check_repository(repo_path))

        if filename:
            self.assertRaises(SCMError,
                              lambda: tool.get_file(filename, HEAD))

        for local_site_name in ('site-1',):
            local_site = LocalSite(name=local_site_name)
            local_site.save()

            repo = Repository(name='SSH Test', path=repo_path, tool=tool_class,
                              local_site=local_site)
            tool = repo.get_scmtool()

            ssh_client = SSHClient(namespace=local_site_name)
            self.assertEqual(ssh_client.storage.get_ssh_dir(),
                             os.path.join(sshdir, local_site_name))
            ssh_client.import_user_key(user_key)
            self.assertEqual(ssh_client.get_user_key(), user_key)

            # Make sure we can verify the repository and access files.
            tool.check_repository(repo_path, local_site_name=local_site_name)

            if filename:
                self.assertNotEqual(tool.get_file(filename, HEAD), None)



from __future__ import unicode_literals

from uuid import uuid4
import os

from django import forms

from reviewboard.attachments.mimetypes import get_uploaded_file_mimetype
from reviewboard.attachments.models import (FileAttachment,
                                            FileAttachmentHistory)
from reviewboard.reviews.models import ReviewRequestDraft


class UploadFileForm(forms.Form):
    """A form that handles uploading of new files."""

    #: The caption for the file.
    caption = forms.CharField(required=False)

    #: The file itself.
    path = forms.FileField(required=True)

    #: An optional file attachment history.
    #:
    #: This is used when creating a new revision for an existing file
    #: attachment. If this is not specified, a new history will be created.
    attachment_history = forms.ModelChoiceField(
        queryset=FileAttachmentHistory.objects.all(),
        required=False)

    def __init__(self, review_request, *args, **kwargs):
        """Initialize the form.

        Args:
            review_request (reviewboard.reviews.models.ReviewRequest):
                The review request to attach the file to.

            args (tuple):
                Extra positional arguments for the form.

            **kwargs (dict):
                Extra keyword arguments for the form.
        """
        super(UploadFileForm, self).__init__(*args, **kwargs)

        self.review_request = review_request

    def clean_attachment_history(self):
        """Validate that the specified file attachment history exists.

        Returns:
            reviewboard.attachments.models.FileAttachmentHistory:
            The history model.
        """
        history = self.cleaned_data['attachment_history']

        if (history is not None and
            not self.review_request.file_attachment_histories.filter(
                pk=history.pk).exists()):
            raise forms.ValidationError(
                'The FileAttachmentHistory provided is not part of this '
                'review request.')

        return history

    def create(self, filediff=None):
        """Create a FileAttachment based on this form.

        Args:
            filediff (reviewboard.diffviewer.models.FileDiff, optional):
                The optional diff to attach this file to (for use when this
                file represents a binary file within the diff).

        Returns:
            reviewboard.attachments.models.FileAttachment:
            The new file attachment model.
        """
        file_obj = self.files['path']
        caption = self.cleaned_data['caption'] or file_obj.name

        mimetype = get_uploaded_file_mimetype(file_obj)
        filename = get_unique_filename(file_obj.name)

        if self.cleaned_data['attachment_history'] is None:
            # This is a new file: create a new FileAttachmentHistory for it
            attachment_history = FileAttachmentHistory()
            attachment_revision = 1

            attachment_history.display_position = \
                FileAttachmentHistory.compute_next_display_position(
                    self.review_request)
            attachment_history.save()
            self.review_request.file_attachment_histories.add(
                attachment_history)
        else:
            attachment_history = self.cleaned_data['attachment_history']

            try:
                latest = attachment_history.file_attachments.latest()
            except FileAttachment.DoesNotExist:
                latest = None

            if latest is None:
                # This should theoretically never happen, but who knows.
                attachment_revision = 1
            elif latest.review_request.exists():
                # This is a new update in the draft.
                attachment_revision = latest.attachment_revision + 1
            else:
                # The most recent revision is part of the same draft. Delete it
                # and replace with the newly uploaded file.
                attachment_revision = latest.attachment_revision
                latest.delete()

        attachment_kwargs = {
            'attachment_history': attachment_history,
            'attachment_revision': attachment_revision,
            'caption': '',
            'draft_caption': caption,
            'orig_filename': os.path.basename(file_obj.name),
            'mimetype': mimetype,
        }

        if filediff:
            file_attachment = FileAttachment.objects.create_from_filediff(
                filediff,
                save=False,
                **attachment_kwargs)
        else:
            file_attachment = FileAttachment(**attachment_kwargs)

        file_attachment.file.save(filename, file_obj, save=True)

        draft = ReviewRequestDraft.create(self.review_request)
        draft.file_attachments.add(file_attachment)
        draft.save()

        return file_attachment


class UploadUserFileForm(forms.Form):
    """A form that handles uploading of user files."""

    #: The caption for the file.
    caption = forms.CharField(required=False)

    #: The file itself.
    path = forms.FileField(required=False)

    def create(self, user, local_site=None):
        """Create a FileAttachment based on this form.

        Args:
            user (django.contrib.auth.models.User):
                The user who owns this file attachment.

            local_site (reviewboard.site.models.LocalSite, optional):
                The optional local site.

        Returns:
            reviewboard.attachments.models.FileAttachment:
            The new file attachment model.
        """
        file_obj = self.files.get('path')

        attachment_kwargs = {
            'uuid': uuid4(),
            'user': user,
            'local_site': local_site,
        }

        if file_obj:
            mimetype = get_uploaded_file_mimetype(file_obj)
            filename = get_unique_filename(file_obj.name)

            attachment_kwargs.update({
                'caption': self.cleaned_data['caption'] or file_obj.name,
                'orig_filename': os.path.basename(file_obj.name),
                'mimetype': mimetype,
            })

            file_attachment = FileAttachment(**attachment_kwargs)
            file_attachment.file.save(filename, file_obj, save=True)
        else:
            attachment_kwargs['caption'] = self.cleaned_data['caption'] or ''

            file_attachment = FileAttachment.objects.create(
                **attachment_kwargs)

        return file_attachment

    def update(self, file_attachment):
        """Update an existing file attachment.

        Args:
            file_attachment (reviewboard.attachments.models.FileAttachment):
                The file attachment to update.

        Returns:
            reviewboard.attachments.models.FileAttachment:
            The updated file attachment.
        """
        caption = self.cleaned_data['caption']
        file_obj = self.files.get('path')

        if caption:
            file_attachment.caption = caption

        if file_obj:
            file_attachment.mimetype = get_uploaded_file_mimetype(file_obj)
            file_attachment.orig_filename = os.path.basename(file_obj.name)
            file_attachment.file.save(get_unique_filename(file_obj.name),
                                      file_obj, save=True)

        file_attachment.save()

        return file_attachment


def get_unique_filename(filename):
    """Return a unique filename.

    Create a unique filename by concatenating a UUID with the given filename.

    Args:
        filename (unicode):
            The original filename.

    Returns:
        unicode:
        A new filename which is more unique.
    """
    return '%s__%s' % (uuid4(), filename)



from __future__ import unicode_literals

from django_evolution.mutations import AddField
from django.db import models


MUTATIONS = [
    AddField('FileAttachment', 'uuid', models.CharField, max_length=255,
             initial=''),
]



"""Review Board-specific specializations of avatar services."""

from __future__ import unicode_literals

from django.utils import six
from django.utils.six.moves.urllib.parse import urlparse
from djblets.avatars.services import \
    FileUploadService as DjbletsFileUploadService

from reviewboard.admin.server import build_server_url


class FileUploadService(DjbletsFileUploadService):
    """A Review Board-specialized file upload avatar services.

    This service is almost identical to the :py:class:`one in Djblets
    <djblets.avatars.services.file_upload.FileUploadService>`, except we ensure
    that all URLs returned are absolute URLs.
    """

    def get_avatar_urls_uncached(self, user, size):
        """Return the avatar URLs for the requested user.

        Args:
            user (django.contrib.auth.models.User):
                The user whose avatar URLs are to be fetched.

            size (int):
                The size (in pixels) the avatar is to be rendered at.

        Returns
            dict:
            A dictionary containing the URLs of the user's avatars at normal-
            and high-DPI.
        """
        urls = super(FileUploadService, self).get_avatar_urls_uncached(user,
                                                                       size)

        return {
            resolution: self._ensure_absolute(url)
            for resolution, url in six.iteritems(urls)
        }

    def _ensure_absolute(self, url):
        """Return the provided URL as an absolute URL.

        Relative URLs will be prefixed with the site's domain method and domain.

        Returns:
            unicode:
            An absolute URL.
        """
        result = urlparse(url)

        if result.scheme and result.netloc:
            return url

        return build_server_url(url)



from __future__ import unicode_literals

import logging

from django import template
from django.template.loader import render_to_string

from reviewboard.extensions.hooks import (CommentDetailDisplayHook,
                                          HeaderActionHook,
                                          HeaderDropdownActionHook,
                                          NavigationBarHook)
from reviewboard.site.urlresolvers import local_site_reverse


register = template.Library()


def action_hooks(context, hook_cls, action_key="action",
                 template_name="extensions/action.html"):
    """Displays all registered action hooks from the specified ActionHook."""
    html = []

    for hook in hook_cls.hooks:
        try:
            for actions in hook.get_actions(context):
                if actions:
                    context.push()
                    context[action_key] = actions

                    try:
                        html.append(render_to_string(template_name, context))
                    except Exception as e:
                        logging.error(
                            'Error when rendering template for action "%s" '
                            'for hook %r in extension "%s": %s',
                            action_key, hook, hook.extension.id, e,
                            exc_info=1)

                    context.pop()
        except Exception as e:
            logging.error('Error when running get_actions() on hook %r '
                          'in extension "%s": %s',
                          hook, hook.extension.id, e, exc_info=1)

    return ''.join(html)


@register.simple_tag(takes_context=True)
def navigation_bar_hooks(context):
    """Displays all registered navigation bar entries."""
    html = []

    for hook in NavigationBarHook.hooks:
        try:
            for nav_info in hook.get_entries(context):
                if nav_info:
                    url_name = nav_info.get('url_name', None)

                    if url_name:
                        nav_info['url'] = local_site_reverse(
                            url_name, request=context.get('request'))

                    context.push()
                    context['entry'] = nav_info
                    html.append(render_to_string(
                        'extensions/navbar_entry.html',
                        context))
                    context.pop()
        except Exception as e:
            extension = hook.extension
            logging.error('Error when running NavigationBarHook.'
                          'get_entries function in extension: "%s": %s',
                          extension.id, e, exc_info=1)

    return ''.join(html)


@register.simple_tag(takes_context=True)
def header_action_hooks(context):
    """Displays all single-entry action hooks for the header bar."""
    return action_hooks(context, HeaderActionHook)


@register.simple_tag(takes_context=True)
def header_dropdown_action_hooks(context):
    """Displays all multi-entry action hooks for the header bar."""
    return action_hooks(context,
                        HeaderDropdownActionHook,
                        "actions",
                        "extensions/header_action_dropdown.html")


@register.simple_tag(takes_context=True)
def comment_detail_display_hook(context, comment, render_mode):
    """Displays all additional detail from CommentDetailDisplayHooks."""
    assert render_mode in ('review', 'text-email', 'html-email')

    html = []

    for hook in CommentDetailDisplayHook.hooks:
        try:
            if render_mode == 'review':
                html.append(hook.render_review_comment_detail(comment))
            elif render_mode in ('text-email', 'html-email'):
                html.append(hook.render_email_comment_detail(
                    comment, render_mode == 'html-email'))
        except Exception as e:
            extension = hook.extension
            logging.error('Error when running CommentDetailDisplayHook with '
                          'render mode "%s" in extension: %s: %s',
                          render_mode, extension.id, e, exc_info=1)

    return ''.join(html)



from __future__ import unicode_literals

import json
import logging
import re

from django import forms
from django.core.cache import cache
from django.core.exceptions import ValidationError
from django.template.loader import render_to_string
from django.utils.six.moves.urllib.error import HTTPError, URLError
from django.utils.six.moves.urllib.parse import quote, quote_plus, urlparse
from django.utils.translation import ugettext_lazy as _, ugettext
from djblets.cache.backend import cache_memoize

from reviewboard.hostingsvcs.errors import (AuthorizationError,
                                            HostingServiceError,
                                            InvalidPlanError,
                                            RepositoryError)
from reviewboard.hostingsvcs.forms import (HostingServiceAuthForm,
                                           HostingServiceForm)
from reviewboard.hostingsvcs.service import HostingService
from reviewboard.scmtools.crypto_utils import (decrypt_password,
                                               encrypt_password)
from reviewboard.scmtools.errors import FileNotFoundError
from reviewboard.scmtools.core import Branch, Commit


class GitLabAPIVersionError(HostingServiceError):
    """Raised if we cannot determine the API version."""

    def __init__(self, message, causes):
        """Initialize the GitLabAPIVersionError.

        Args:
            message (unicode):
                The exception message.

            causes (list of Exception):
                The underlying exceptions.
        """
        super(GitLabAPIVersionError, self).__init__(message)

        self.causes = causes

    def __repr__(self):
        """Return a representation of the exception.

        Returns:
            unicode:
            A representation of the exception.
        """
        return ('<GitLabAPIVersionError(message=%r, causes=%r)>'
                % (self.message, self.causes))


class GitLabHostingURLWidget(forms.Widget):
    """A custom input widget for selecting a GitLab host.

    The user can choose between gitlab.com-hosted and self-hosted instances of
    GitLab.
    """

    GITLAB = 'https://gitlab.com'
    CUSTOM = 'custom'

    CHOICES = (
        (GITLAB, _('gitlab.com')),
        (CUSTOM, _('Custom')),
    )

    def value_from_datadict(self, data, files, name):
        """Extract the value from the form data.

        Args:
            data (dict):
                The form data.

            files (dict):
                The files.

            name (unicode):
                The name of the form field.

        Returns:
            unicode:
            The form value.
        """
        if data:
            return data.get(name)

        return self.GITLAB

    def render(self, name, value, attrs=None):
        """Render the widget.

        Args:
            name (unicode):
                The name of the widget.

            value (unicode):
                The value of the widget.

            attrs (dict, optional):
                Additional attributes to pass to the widget.

        Returns:
            django.util.safestring.SafeText:
            The rendered widget.
        """
        attrs = self.build_attrs(attrs)

        return render_to_string('hostingsvcs/gitlab/url_widget.html', {
            'attrs': attrs,
            'id': attrs.pop('id'),
            'is_custom': value and value != self.GITLAB,
            'name': name,
            'value': value or '',
        })


class GitLabAuthForm(HostingServiceAuthForm):
    """An authentication form for the GitLab hosting service.

    This form allows user to select between gitlab.com and self-hosted
    instances of GitLab.
    """

    hosting_url = forms.CharField(
        label=_('Service URL'),
        required=True,
        widget=GitLabHostingURLWidget(attrs={'size': 30}))

    private_token = forms.CharField(
        label=_('API Token'),
        required=True,
        help_text=_(
            'Your GitLab API token. In newer versions of GitLab, you can '
            'create one under <code>Settings &gt; Access Tokens &gt; Personal '
            'Access</code>. This token will need the <code>api</code> scope. '
            'In older versions of GitLab, you can find this under '
            '<code>Profile Settings &gt; Account &gt; Private Token</code>.'
        )
    )

    def __init__(self, *args, **kwargs):
        """Initialize the GitLabAuthForm.

        Args:
            *args (tuple):
                Positional arguments to pass to the base class constructor.

            **kwargs (dict):
                Keyword arguments to pass to the base class constructor.
        """
        super(GitLabAuthForm, self).__init__(*args, **kwargs)

        del self.fields['hosting_account_password']

    def clean_hosting_url(self):
        """Clean the hosting_url field.

        This method ensures that the URL has a scheme.

        Returns:
            unicode: The URL.

        Raises:
            django.core.exceptions.ValidationError:
                The URL was missing a scheme.
        """
        hosting_url = self.cleaned_data['hosting_url']
        result = urlparse(hosting_url)

        if not result.scheme:
            raise ValidationError(
                _('Invalid hosting URL "%(url)s": missing scheme (e.g., HTTP '
                  'or HTTPS)')
                % {
                    'url': hosting_url,
                }
            )

        return hosting_url

    def get_credentials(self):
        """Return the credentials for the form.

        Returns:
            dict:
            A dict containing the values of the ``username`` and
            ``private_token`` fields.
        """
        credentials = {
            'username': self.cleaned_data['hosting_account_username'],
            'private_token': self.cleaned_data['private_token'],
        }

        two_factor_auth_code = \
            self.cleaned_data.get('hosting_account_two_factor_auth_code')

        if two_factor_auth_code:
            credentials['two_factor_auth_code'] = two_factor_auth_code

        return credentials


class GitLabPersonalForm(HostingServiceForm):
    gitlab_personal_repo_name = forms.CharField(
        label=_('Repository name'),
        max_length=64,
        required=True,
        widget=forms.TextInput(attrs={'size': '60'}))


class GitLabGroupForm(HostingServiceForm):
    gitlab_group_name = forms.CharField(
        label=_('GitLab group name'),
        max_length=64,
        required=True,
        widget=forms.TextInput(attrs={'size': '60'}))

    gitlab_group_repo_name = forms.CharField(
        label=_('Repository name'),
        max_length=64,
        required=True,
        widget=forms.TextInput(attrs={'size': '60'}))


class GitLab(HostingService):
    """Hosting service support for GitLab.

    GitLab is a self-installed source hosting service that supports Git
    repositories. It's available at https://gitlab.org/.
    """
    name = 'GitLab'

    # The maximum number of commits returned from each call to get_commits()
    COMMITS_PER_PAGE = 20

    self_hosted = True
    needs_authorization = True
    supports_bug_trackers = True
    supports_post_commit = True
    supports_repositories = True
    supported_scmtools = ['Git']

    # Pagination links (in GitLab 6.8.0+) take the form:
    # '<http://gitlab/api/v3/projects?page=2&per_page=100>; rel="next"'
    LINK_HEADER_RE = re.compile(r'\<(?P<url>[^\>]+)\>; rel="next"')

    auth_form = GitLabAuthForm

    plans = [
        ('personal', {
            'name': _('Personal'),
            'form': GitLabPersonalForm,
            'repository_fields': {
                'Git': {
                    'path': 'git@%(hosting_domain)s:'
                            '%(hosting_account_username)s/'
                            '%(gitlab_personal_repo_name)s.git',
                    'mirror_path': '%(hosting_url)s/'
                                   '%(hosting_account_username)s/'
                                   '%(gitlab_personal_repo_name)s.git',
                },
            },
            'bug_tracker_field': '%(hosting_url)s/'
                                 '%(hosting_account_username)s/'
                                 '%(gitlab_personal_repo_name)s/issues/%%s'
        }),
        ('group', {
            'name': _('Group'),
            'form': GitLabGroupForm,
            'repository_fields': {
                'Git': {
                    'path': 'git@%(hosting_domain)s:'
                            '%(gitlab_group_name)s/'
                            '%(gitlab_group_repo_name)s.git',
                    'mirror_path': '%(hosting_url)s/%(gitlab_group_name)s/'
                                   '%(gitlab_group_repo_name)s.git',
                },
            },
            'bug_tracker_field': '%(hosting_url)s/%(gitlab_group_name)s/'
                                 '%(gitlab_group_repo_name)s/issues/%%s'
        }),
    ]

    def check_repository(self, plan=None, *args, **kwargs):
        """Checks the validity of a repository.

        This will perform an API request against GitLab to get
        information on the repository. This will throw an exception if
        the repository was not found, and return cleanly if it was found.
        """
        self._find_repository_id(
            plan,
            self._get_repository_owner(plan, kwargs),
            self._get_repository_name(plan, kwargs))

    def authorize(self, username, credentials, hosting_url, *args, **kwargs):
        """Authorize the GitLab repository.

        GitLab uses HTTP Basic Auth for the API, so this will store the
        provided password, encrypted, for use in later API requests.

        Args:
            username (unicode):
                The username of the account being linked.

            credentials (dict):
                Authentication credentials.

            hosting_url (unicode):
                The URL of the GitLab server.

            *args (tuple, unused):
                Ignored positional arguments.

            **kwargs (dict, unused):
                Ignored keyword arguments.

        Raises:
            reviewboard.hostingsvcs.errors.AuthorizationError:
                Authorization could not be completed successfully.

            reviewboard.hostingsvcs.errors.HostingServiceError:
                An HTTP or other unexpected error occurred.
        """
        # This will raise an exception if it fails, which the form will
        # catch.
        try:
            self._try_api_versions(
                hosting_url,
                path='/projects?per_page=1',
                headers={
                    b'PRIVATE-TOKEN':
                        credentials['private_token'].encode('utf-8'),
                })
        except AuthorizationError:
            raise
        except HTTPError as e:
            if e.code == 404:
                raise HostingServiceError(
                    ugettext('A GitLab server was not found at the '
                             'provided URL.'))
            else:
                logging.exception('Unexpected HTTP error when linking GitLab '
                                  'account for %s: %s',
                                  username, e)
                raise HostingServiceError(
                    ugettext('Unexpected HTTP error %s.')
                    % e.code)
        except Exception as e:
            logging.exception('Unexpected error when linking GitLab account '
                              'for %s: %s',
                              username, e)
            raise HostingServiceError(
                ugettext('Unexpected error "%s"') % e)

        self.account.data['private_token'] = \
            encrypt_password(credentials['private_token'])
        self.account.save()

    def is_authorized(self):
        """Determine if the account has supported authorization tokens.

        This checks if we have previously stored a private token for the
        account. It does not validate that the token still works.

        Returns:
            bool:
            Whether or not the account is authorized with GitLab.
        """
        return 'private_token' in self.account.data

    def get_file(self, repository, path, revision, base_commit_id=None,
                 *args, **kwargs):
        """Fetch a file from GitLab.

        This will perform an API request to fetch the contents of a file.

        Args:
            repository (reviewboard.scmtools.models.Repository):
                The repository to fetch the file from.

            path (unicode):
                The file path.

            revision (unicode):
                The SHA1 of the file blob.

            base_commit_id (unicode, optional):
                An optional commit SHA1.

            *args (tuple, unused):
                Ignored positional arguments.

            **kwargs (dict, unused):
                Ignored keyword arguments.

        Returns:
            bytes:
            The file data at the requested revision.

        Raises:
            reviewboard.scmtools.errors.FileNotFoundError:
                The file could not be retrieved.
        """
        try:
            data, headers = self._api_get(
                repository.hosting_account.hosting_url,
                self._get_blob_url(repository, path, revision, base_commit_id),
                raw_content=True)
            return data
        except (HTTPError, URLError):
            raise FileNotFoundError(path, revision)

    def get_file_exists(self, repository, path, revision, base_commit_id=None,
                        *args, **kwargs):
        """Determine if a file exists.

        This will perform an API request to fetch the metadata for a file.

        Args:
            repository (reviewboard.scmtools.models.Repository):
                The repository to fetch the file from.

            path (unicode):
                The file path.

            revision (unicode):
                The SHA1 of the file blob.

            base_commit_id (unicode, optional):
                An optional commit SHA1.

            *args (tuple, unused):
                Ignored positional arguments.

            **kwargs (dict, unused):
                Ignored keyword arguments.

        Returns:
            bool:
            Whether or not the file exists.

        Raises:
            reviewboard.scmtools.errors.AuthorizationError:
                There was an issue with the authorization credentials.

            reviewboard.scmtools.errors.FileNotFoundError:
                The file could not be retrieved.
        """
        try:
            self._api_get(
                repository.hosting_account.hosting_url,
                self._get_blob_url(repository, path, revision, base_commit_id),
                raw_content=True)

            return True
        except (HTTPError, URLError):
            return False

    def get_branches(self, repository):
        """Return a list of branches.

        This will perform an API request to fetch a list of branches.

        Args:
            repository (reviewboard.scmtools.models.Repository):
                The repository to get branches from.

        Returns:
            list of reviewboard.scmtools.core.Branch:
            The branches available.

        Raises:
            reviewboard.scmtools.errors.AuthorizationError:
                There was an issue with the authorization credentials.

            urllib2.HTTPError:
                There was an error communicating with the server.
        """
        repo_api_url = ('%s/repository/branches'
                        % self._get_repo_api_url(repository))
        refs = self._api_get(repository.hosting_account.hosting_url,
                             repo_api_url)[0]

        results = []

        for ref in refs:
            if 'name' in ref:
                name = ref['name']
                results.append(Branch(id=name,
                                      commit=ref['commit']['id'],
                                      default=(name == 'master')))

        return results

    def get_commits(self, repository, branch=None, start=None):
        """Return a list of commits

        This will perform an API request to fetch a list of commits.

        Args:
            repository (reviewboard.scmtools.models.Repository):
                The repository to fetch commits from.

            branch (unicode, optional):
                The branch to fetch commits from. If not provided, the default
                branch will be used.

            start (unicode, optional):
                The commit to start fetching form.

                If provided, this argument will override ``branch``. Otherwise,
                if neither are provided, the default branch will be used.

        Returns:
            list of reviewboard.scmtools.core.Commit:
            The commits from the API.

        Raises:
            reviewboard.scmtools.errors.AuthorizationError:
                There was an issue with the authorization credentials.

            urllib2.HTTPError:
                There was an error communicating with the server.
        """
        # Ask GitLab for 21 commits per page. GitLab's API doesn't
        # include the parent IDs, so we use subsequent commits to fill
        # them in (allowing us to return 20 commits with parent IDs).
        page_size = self.COMMITS_PER_PAGE + 1

        repo_api_url = ('%s/repository/commits?per_page=%s'
                        % (self._get_repo_api_url(repository),
                           page_size))

        if start:
            # If start parameter is given, use it as the latest commit to log
            # from, so that we fetch a page of commits, and the first commit id
            # on the page is the start parameter.
            repo_api_url += '&ref_name=%s' % start
        elif branch:
            # The branch is optional. If it is not given, use the default
            # branch. The default branch is set to 'master' in get_branches()
            repo_api_url += '&ref_name=%s' % branch

        # The GitLab API will return a tuple consists of two elements.
        # The first one is a list of commits, and the other one is an instance
        # type object containing all kinds of headers, which is not required.
        commits = self._api_get(repository.hosting_account.hosting_url,
                                repo_api_url)[0]

        results = []

        for idx, item in enumerate(commits):
            commit = self._parse_commit(item)

            if idx > 0:
                # Note that GitLab API documents do not show any returned
                # 'parent_id' from the query for a list of commits. So we use
                # the current commit id as the previous commit's parent id, and
                # remove the last commit from results.
                results[idx - 1].parent = commit.id

            results.append(commit)

        # Strip off the last commit since we don't know its parent id yet.
        if len(commits) == page_size:
            results.pop()

        return results

    def get_change(self, repository, revision):
        """Fetch a single commit from GitLab.

        Args:
            repository (reviewboard.scmtools.models.Repository):
                The repository in question.

            revision (unicode):
                The SHA1 hash of the commit to fetch.

        Returns:
            reviewboard.scmtools.core.Commit:
            The commit in question.

        Raises:
            reviewboard.scmtools.errors.AuthorizationError:
                There was an issue with the authorization credentials.

            urllib2.HTTPError:
                There was an error communicating with the server.
        """
        repo_api_url = self._get_repo_api_url(repository)
        private_token = self._get_private_token()

        # Step 1: Fetch the commit itself that we want to review, to get
        # the parent SHA and the commit message. Hopefully this information
        # is still in cache so we don't have to fetch it again. However, the
        # parent SHA is probably empty.
        commit = cache.get(repository.get_commit_cache_key(revision))

        if commit is None:
            commit_api_url = ('%s/repository/commits/%s'
                              % (repo_api_url, revision))

            # This response from GitLab consists of one dict type commit and
            # on instance type header object. Only the first element is needed.
            commit_data = self._api_get(repository.hosting_account.hosting_url,
                                        commit_api_url)[0]

            commit = self._parse_commit(commit_data)
            commit.parent = commit_data['parent_ids'][0]

        # Step 2: Get the diff. The revision is the commit header in here.
        # Firstly, a diff url should be built up, which has the format of
        # <hosting_url>/<user-name>/<project-name>/commit/<revision>.diff,
        # then append the private_token to the end of the url and get the diff.

        hosting_url = self.account.hosting_url

        if not hosting_url.endswith('/'):
            hosting_url += '/'

        # Get the project path with the namespace.
        path_api_url = ('%s?private_token=%s'
                        % (repo_api_url, private_token))
        project = self._api_get(repository.hosting_account.hosting_url,
                                path_api_url)[0]
        path_with_namespace = project['path_with_namespace']

        # Build up diff url and get diff.
        diff_url = ('%s%s/commit/%s.diff?private_token=%s'
                    % (hosting_url, path_with_namespace, revision,
                       private_token))
        diff, headers = self.client.http_get(
            diff_url,
            headers={'Accept': 'text/plain'})

        # Remove the last two lines. The last line is 'libgit <version>',
        # and the second last line is '--', ending with '\n'. To avoid the
        # error from parsing the empty file (size is 0), split the string into
        # two parts using the delimiter '--\nlibgit'. If only use '\n' or '--'
        # delimiter, more characters might be stripped out from file
        # modification commit diff.
        diff = diff.rsplit(b'--\nlibgit', 2)[0]

        # Make sure there's a trailing newline.
        if not diff.endswith(b'\n'):
            diff += b'\n'

        commit.diff = diff
        return commit

    def _find_repository_id(self, plan, owner, repo_name):
        """Find the ID of a repository matching the given name and owner.

        If the repository could not be found, an appropriate error will be
        raised.

        Args:
            plan (unicode):
                The plan name.

            owner (unicode):
                The name of the owning group or user.

            repo_name (unicode):
                The name of the repository.

        Returns:
            int:
            The ID of the repository.

        Raises:
            reviewboard.scmtools.errors.AuthorizationError:
                There was an issue with the authorization credentials.

            reviewboard.scmtools.errors.RepositoryError:
                The repository could be found or accessed.

            urllib2.HTTPError:
                There was an error communicating with the server.
        """
        if self._get_api_version(self.account.hosting_url) == '3':
            return self._find_repository_id_v3(plan, owner, repo_name)
        else:
            return self._find_repository_id_v4(plan, owner, repo_name)

    def _find_repository_id_v4(self, plan, owner, repo_name):
        """Find the ID of a repository matching the given name and owner.

        If the repository could not be found, an appropriate error will be
        raised.

        Args:
            plan (unicode):
                The plan name.

            owner (unicode):
                The name of the owning group or user.

            repo_name (unicode):
                The name of the repository.

        Returns:
            int:
            The ID of the repository.

        Raises:
            reviewboard.scmtools.errors.AuthorizationError:
                There was an issue with the authorization credentials.

            reviewboard.scmtools.errors.RepositoryError:
                The repository could be found or accessed.

            urllib2.HTTPError:
                There was an error communicating with the server.
        """
        project = '%s/%s' % (owner, repo_name)

        try:
            data, headers = self._api_get(self.account.hosting_url,
                                          'projects/%s' % quote_plus(project))
            return data['id']
        except HTTPError as e:
            if e.code == 404:
                raise RepositoryError(
                    ugettext('A repository with this name was not found, or '
                             'your user may not own it.'))

            raise

    def _find_repository_id_v3(self, plan, owner, repo_name):
        """Find the ID of a repository matching the given name and owner.

        If the repository could not be found, an appropriate error will be
        raised.

        Args:
            plan (unicode):
                The plan name.

            owner (unicode):
                The name of the owning group or user.

            repo_name (unicode):
                The name of the repository.

        Returns:
            int:
            The ID of the repository.

        Raises:
            reviewboard.scmtools.errors.AuthorizationError:
                There was an issue with the authorization credentials.

            reviewboard.scmtools.errors.RepositoryError:
                The repository could be found or accessed.

            urllib2.HTTPError:
                There was an error communicating with the server.
        """
        # GitLab claims pagination support, but it has a number of problems.
        # We have no idea how many pages there are, or even if there's another
        # page of items. Furthermore, if we try to go beyond the last page,
        # we just get the first again, so we can't attempt to guess very
        # well.
        #
        # If the list doesn't return the repository, the user is out of luck.
        #
        # This is true as of GitLab 6.4.3.
        if plan == 'personal':
            repositories = self._api_get_repositories()

            for repository_entry in repositories:
                namespace = repository_entry['namespace']

                if (namespace['path'] == owner and
                    repository_entry['path'] == repo_name):
                    # This is the repository we wanted to find.
                    return repository_entry['id']

            raise RepositoryError(
                ugettext('A repository with this name was not found, or your '
                         'user may not own it.'))
        elif plan == 'group':
            groups = self._api_get_groups()

            for group_entry in groups:
                # If the full path is available, use that (to support nested
                # groups). Otherwise fall back on the group name.
                group_name = group_entry.get('full_path', group_entry['name'])

                if group_name == owner:
                    group_id = group_entry['id']
                    group_data = self._api_get_group(group_id)
                    repositories = group_data['projects']

                    for repository_entry in repositories:
                        if repository_entry['name'] == repo_name:
                            return repository_entry['id']

                    raise RepositoryError(
                        ugettext('A repository with this name was not '
                                 'found on this group, or your user may '
                                 'not have access to it.'))
            raise RepositoryError(
                ugettext('A group with this name was not found, or your user '
                         'may not have access to it.'))
        else:
            raise InvalidPlanError(plan)

    def _api_get_group(self, group_id):
        """Return information about a given group.

        Args:
            group_id (int):
                The ID of the group to fetch repositories for.

        Returns:
            dict:
            Information about the requested group.

        Raises:
            reviewboard.scmtools.errors.AuthorizationError:
                There was an issue with the authorization credentials.

            urllib2.HTTPError:
                There was an error communicating with the server.
        """
        return self._api_get(self.account.hosting_url,
                             'groups/%s' % group_id)[0]

    def _api_get_groups(self):
        """Return a list of groups the user has access to.

        This will fetch up to 100 groups from GitLab. These are all groups the
        user has any form of access to.

        Returns:
            list of dict:
            The list of group information.

        Raises:
            reviewboard.scmtools.errors.AuthorizationError:
                There was an issue with the authorization credentials.

            urllib2.HTTPError:
                There was an error communicating with the server.
        """
        return self._api_get(self.account.hosting_url,
                             'groups?per_page=100')[0]

    def _api_get_repositories(self):
        """Return a list of repositories the user has access to.

        These are all repositories the user has any form of access to.

        Returns:
            list of dict:
            A list of the parsed JSON responses.

        Raises:
            reviewboard.scmtools.errors.AuthorizationError:
                There was an issue with the authorization credentials.

            urllib2.HTTPError:
                There was an error communicating with the server.
        """
        return self._api_get_list(self.account.hosting_url,
                                  'projects?per_page=100')

    def _build_api_url(self, hosting_url, path, api_version=None):
        """Build an API URL.

        Args:
            hosting_url (unicode):
                The URL of the GitLab server.

            path (unicode):
                The API path (not including :samp:`/api/v{version}/`) to build
                the URL for.

            api_version (int, optional):
                The version of the API (3 or 4) to build the URL for.

                If not provided, it will be determined via the cache or, if
                uncached, from the server itself.

        Returns:
            unicode:
            The URL.
        """
        if api_version is None:
            api_version = self._get_api_version(hosting_url)

        return '%s/api/v%s/%s' % (hosting_url.rstrip('/'), api_version,
                                  path.lstrip('/'))

    def _get_blob_url(self, repository, path, revision, base_commit_id=None):
        """Return the URL for accessing the contents of a file.

        Args:
            repository (reviewboard.scmtools.models.Repository):
                The repository.

            path (unicode):
                The path to the file.

            revision (unicode):
                The SHA1 of the blob.

            base_commit_id (unicode, optional):
                The SHA1 of the commit to fetch the file at.

                If provided, this will use their standard blob API, which takes
                a commit ID and a file path.

                If not provided, it will try the newer API for accessing based
                on a blob SHA1. This requires a new enough version of GitLab,
                which we unfortunately cannot detect through their API.

        Returns:
            unicode:
            The blob URL.
        """
        repo_api_url = self._get_repo_api_url(repository)
        api_version = self._get_api_version(self.account.hosting_url)

        if api_version == '3':
            # Not all versions of GitLab support a blob ID, so if a base commit
            # ID is provided, we're going to use that instead.
            if base_commit_id:
                return ('%s/repository/blobs/%s?filepath=%s'
                        % (repo_api_url, base_commit_id, quote(path)))
            else:
                return ('%s/repository/raw_blobs/%s'
                        % (repo_api_url, revision))
        else:
            return ('%s/repository/blobs/%s/raw'
                    % (repo_api_url, revision))

    def _get_repo_api_url(self, repository):
        """Return the base URL for a repository's API.

        The first time this is called, it will look up the repository ID
        through the API. This may take time, but only has to be done once
        per repository.

        Args:
            repository (reviewboard.scmtools.models.Repository):
                The repository.

        Returns:
            unicode:
            The URL of the repository.
        """
        return 'projects/%s' % self._get_repository_id(repository)

    def _get_repository_id(self, repository):
        """Return the ID of a repository.

        If the ID is unknown, this will attempt to look up the ID in the
        list of repositories the user has access to. It will then store the
        ID for later requests, to prevent further lookups.

        Args:
            repository (reviewboard.scmtools.models.Repository):
                The repository.

        Returns:
            int:
            The ID of the repository in GitLab.
        """
        key = 'gitlab_project_id'

        if key not in repository.extra_data:
            plan = repository.extra_data['repository_plan']

            repository.extra_data[key] = self._find_repository_id(
                plan,
                self._get_repository_owner(plan, repository.extra_data),
                self._get_repository_name(plan, repository.extra_data))
            repository.save()

        return repository.extra_data[key]

    def _get_repository_owner(self, plan, extra_data):
        """Return the owner of a repository.

        Args:
            plan (unicode):
                The plan name. This should be one of either ``'personal'`` or
                ``'group'``.

            extra_data (dict):
                The
                :py:attr:`~reviewboard.scmtools.models.Repository.extra_data`
                attribute.

        Returns:
            unicode:
            The owner of the repository.

            If this is a personal repository, the owner will be the user who
            has linked their account to GitLab.

            If this is a group repository, the owner will be the group name.

        Raises:
              reviewboard.hostingsvcs.errors.InvalidPlanError:
                  Raised when the plan is not a valid choice.
        """
        if plan == 'personal':
            return self.account.username
        elif plan == 'group':
            return extra_data['gitlab_group_name']
        else:
            raise InvalidPlanError(plan)

    def _get_repository_name(self, plan, extra_data):
        """Return the name of the repository.

        Args:
            plan (unicode):
                The repository plan.

            extra_data (dict):
                The ``extra_data`` attribute of the corresponding
                :py:class:`reviewboard.scmtools.models.Repository`.

        Returns:
            unicode:
            The name of the plan.

        Raises:
            reviewboard.hostingsvcs.errors.InvalidPlanError:
                An invalid plan was given.
        """
        if plan == 'personal':
            return extra_data['gitlab_personal_repo_name']
        elif plan == 'group':
            return extra_data['gitlab_group_repo_name']
        else:
            raise InvalidPlanError(plan)

    def _get_private_token(self):
        """Return the private token used for authentication.

        Returns:
            unicode:
            The API token.
        """
        return decrypt_password(self.account.data['private_token'])

    def _api_get(self, hosting_url=None, path=None, url=None,
                 raw_content=False):
        """Make a request to the GitLab API and return the result.

        If ``hosting_url`` and ``path`` are provided, the API version will be
        deduced from the server. Otherwise, the full URL given in ``url`` will
        be used.

        Args:
            hosting_url (unicode, optional):
                The host of the repository.

            path (unicode, optional):
                The path after :samp:`/api/v{version}`.

            url (unicode, optional):
                If provided, the full URL to retrieve. Passing ``hosting_url``
                and ``path`` should be preferred over this argument.

            raw_content (bool, optional):
                Whether or not to return the raw content (if ``True``) or to
                parse it as JSON (if ``False``).

                Defaults to ``False``.

        Returns:
            object:
            The response.

        Raises:
            reviewboard.scmtools.errors.AuthorizationError:
                There was an issue with the authorization credentials.

            urllib2.HTTPError:
                There was an error communicating with the server.
        """
        if url:
            assert not hosting_url
            assert not path
        else:
            url = self._build_api_url(hosting_url, path)

        headers = {
            b'PRIVATE-TOKEN': self._get_private_token(),
        }

        if not raw_content:
            headers[b'Accept'] = b'application/json'

        try:
            data, headers = self.client.http_get(url, headers)

            if raw_content:
                return data, headers
            else:
                return json.loads(data), headers
        except HTTPError as e:
            if e.code == 401:
                raise AuthorizationError(
                    ugettext('The login or password is incorrect.'))

            raise

    def _api_get_list(self, hosting_url, path):
        """Make a request to a GitLab list API and return the full list.

        If the server provides a "next" link in the headers (GitLab 6.8.0+),
        this will follow that link and fetch all the results. Otherwise, this
        will provide only the first page of results.

        Args:
            hosting_url (unicode):
                The GitLab server URL.

            path (unicode):
                The path to list resource to fetch.

        Returns:
            list of dict:
            The list of all objects retrieved from the path and its subsequent
            pages.

        Raises:
            reviewboard.scmtools.errors.AuthorizationError:
                There was an issue with the authorization credentials.

            urllib2.HTTPError:
                There was an error communicating with the server.
        """
        all_data = []
        url = self._build_api_url(hosting_url, path)

        while url:
            data, headers = self._api_get(url=url)

            all_data += data

            url = None
            for link in headers.get('link', '').split(', '):
                m = self.LINK_HEADER_RE.match(link)
                if m:
                    url = m.group('url')
                    break

        return all_data

    def _get_api_version(self, hosting_url):
        """Return the version of the API supported by the given server.

        This method will cache the result.

        Args:
            hosting_url (unicode):
                The URL of the GitLab server.

        Returns:
            unicode:
            The version of the API as a string.

            It is returned as a string because
            :py:func:`djblets.cache.backend.cache_memoize` does not work on
            integer results.
        """
        headers = {}

        if self.account.data and 'private_token' in self.account.data:
            headers[b'PRIVATE-TOKEN'] = decrypt_password(
                self.account.data['private_token'])

        return cache_memoize(
            'gitlab-api-version:%s' % hosting_url,
            expiration=3600,
            lookup_callable=lambda: self._try_api_versions(
                hosting_url,
                headers=headers,
                path='/projects?per_page=1',
            )[0])

    def _try_api_versions(self, hosting_url, path, http_method='get',
                          use_json=False, **request_kwargs):
        """Try different API versions and return the first valid response.

        Args:
            hosting_url (unicode):
                The URL of the GitLab server.

            path (unicode):
                The API path to retrieve, not including
                :samp:`/api/v{version}`.

            http_method (unicode, optional):
                The method to use. Defaults to ``GET``.

            use_json (bool, optional):
                Whether or not to interpret the results as JSON.

            **request_kwargs (dict):
                Additional keyword arguments to pass to the request method.

        Returns:
            tuple:
            A 3-tuple of:

            * The API version (:py:class:`unicode`).
            * The response body (:py:class:`bytes` or :py:class:`dict`).
            * The response headers (:py:class:`dict`).

        Raises:
            reviewboard.scmtools.errors.AuthorizationError:
                There was an issue with the authorization credentials.

            GitLabAPIVersionError:
                The API version could be determined.
        """
        http_method = http_method.lower()

        if use_json:
            method = getattr(self.client, 'json_%s' % http_method)
        else:
            method = getattr(self.client, 'http_%s' % http_method)

        errors = []

        for api_version in ('4', '3'):
            url = self._build_api_url(hosting_url, path,
                                      api_version=api_version)

            try:
                rsp, headers = method(url, **request_kwargs)
            except HTTPError as e:
                if e.code == 401:
                    raise AuthorizationError('The API token is invalid.')

                errors.append(e)
            except Exception as e:
                errors.append(e)
            else:
                return api_version, rsp, headers

        raise GitLabAPIVersionError(
            'Could not determine GitLab API version for "%s"'
            % hosting_url,
            errors,
        )

    def _parse_commit(self, commit_data):
        """Return a Commit object based on data return from the API.

        Args:
            commit_data (dict):
                The data returned from the GitLab API.

        Returns
            reviewboard.scmtools.core.Commit:
            The parsed commit.
        """
        # Older versions of GitLab don't have a ``message`` field and
        # instead only offer the ``title`` (i.e., the first line of the
        # commit message).
        return Commit(
            author_name=commit_data['author_name'],
            id=commit_data['id'],
            date=commit_data['created_at'],
            message=commit_data.get('message', commit_data.get('title', '')))



from __future__ import unicode_literals

import logging
from warnings import warn

from django.core.urlresolvers import reverse
from django.utils.translation import ugettext_lazy as _
from djblets.configforms.mixins import DynamicConfigPageMixin
from djblets.configforms.pages import ConfigPage
from djblets.configforms.registry import ConfigPageRegistry
from djblets.registries.errors import ItemLookupError
from djblets.registries.mixins import ExceptionFreeGetterMixin

from reviewboard.admin.server import build_server_url
from reviewboard.accounts.forms.pages import (AccountSettingsForm,
                                              APITokensForm,
                                              AvatarSettingsForm,
                                              ChangePasswordForm,
                                              PrivacyForm,
                                              GroupsForm,
                                              OAuthApplicationsForm,
                                              OAuthTokensForm,
                                              ProfileForm)
from reviewboard.deprecation import RemovedInReviewBoard40Warning


class AccountPageRegistry(ExceptionFreeGetterMixin, ConfigPageRegistry):
    """A registry for managing account pages."""

    lookup_attrs = ('page_id',)

    def get_defaults(self):
        """Return the default page classes.

        Returns:
            type: The page classes, as subclasses of :py:class:`AccountPage`.
        """
        return (PrivacyPage, ProfilePage, AccountSettingsPage, GroupsPage,
                AuthenticationPage, OAuth2Page)

    def unregister(self, page_class):
        """Unregister the page class.

        Args:
            page_class (type):
                The page class to unregister.

        Raises:
            ItemLookupError:
                This exception is raised if the specified page class cannot
                be found.
        """
        try:
            super(AccountPageRegistry, self).unregister(page_class)
        except ItemLookupError as e:
            logging.error(e)
            raise e


class AccountPage(DynamicConfigPageMixin, ConfigPage):
    """Base class for a page of forms in the My Account page.

    Each AccountPage is represented in the My Account page by an entry
    in the navigation sidebar. When the user has navigated to that page,
    any forms shown on the page will be displayed.

    Extensions can provide custom pages in order to offer per-user
    customization.
    """

    registry = AccountPageRegistry()

    @classmethod
    def get_absolute_url(cls):
        """Return the absolute URL of the page.

        Returns:
            unicode:
            The absolute URL of the page.
        """
        assert cls.page_id
        return (
            '%s#%s'
            % (build_server_url(reverse('user-preferences')),
               cls.page_id)
        )


class AccountSettingsPage(AccountPage):
    """A page containing the primary settings the user can customize."""

    page_id = 'settings'
    page_title = _('Settings')
    form_classes = [AccountSettingsForm]


class AuthenticationPage(AccountPage):
    """A page containing authentication-related forms.

    By default, this just shows the Change Password form, but extensions
    can provide additional forms for display.
    """

    page_id = 'authentication'
    page_title = _('Authentication')
    form_classes = [ChangePasswordForm, APITokensForm, OAuthTokensForm]


class ProfilePage(AccountPage):
    """A page containing settings for the user's profile."""

    page_id = 'profile'
    page_title = _('Profile')
    form_classes = [ProfileForm, AvatarSettingsForm]


class GroupsPage(AccountPage):
    """A page containing a filterable list of groups to join."""

    page_id = 'groups'
    page_title = _('Groups')
    form_classes = [GroupsForm]


class OAuth2Page(AccountPage):
    """A page containing a list of OAuth2 applications to manage."""

    page_id = 'oauth2'
    page_title = 'OAuth2 Applications'
    form_classes = [OAuthApplicationsForm]


class PrivacyPage(AccountPage):
    """A page containing information on a user's privacy rights."""

    page_id = 'privacy'
    page_title = _('My Privacy Rights')
    form_classes = [PrivacyForm]


def register_account_page_class(cls):
    """Register a custom account page class.

    A page ID is considered unique and can only be registered once.

    Args:
        cls (type):
            The page class to register, as a subclass of
            :py:class:`AccountPage`.

    Raises:
        djblets.registries.errors.AlreadyRegisteredError:
            Raised if the page or any of its forms have already been
            registered.

        djblets.registries.errors.RegistrationError:
            Raised if the page shares an attribute with an already
            registered page or if any of its forms share an attribute
            with an already registered form.
    """
    warn('register_account_page_class is deprecated in Review Board 3.0 and '
         'will be removed; use AccountPage.registry.register instead.',
         RemovedInReviewBoard40Warning)
    AccountPage.registry.register(cls)


def unregister_account_page_class(page_cls):
    """Unregister a previously registered account page class.

    Args:
        page_cls (type):
            The page class to unregister, as a subclass of
            :py:class:`AccountPage`.
    """
    warn('unregister_account_page_class is deprecated in Review Board 3.0 and '
         'will be removed; use AccountPage.registry.unregister instead.',
         RemovedInReviewBoard40Warning)
    AccountPage.registry.unregister(page_cls)


def get_page_class(page_id):
    """Return the account page class with the specified ID.

    Args:
        page_id (unicode):
            The page's unique identifier.

    Returns:
        type:
        The :py:class:`AccountPage` subclass, or ``None`` if it could not be
        found.
    """
    warn('get_page_class is deprecated in Review Board 3.0 and will be '
         'removed; use AccountPage.registry.get instead.',
         RemovedInReviewBoard40Warning)
    return AccountPage.registry.get('page_id', page_id)


def get_page_classes():
    """Yield all registered page classes.

    Yields:
        type: Each registered page class, as a subclass of
        :py:class:`AccountPage`.
    """
    warn('get_page_classes is deprecated in Review Board 3.0 and will be '
         'removed; iterate through AccountPage.registry instead.',
         RemovedInReviewBoard40Warning)
    return iter(AccountPage.registry)



from __future__ import unicode_literals

from django.contrib.auth.models import User
from django.test import RequestFactory
from djblets.util.templatetags.djblets_images import crop_image
from kgb import SpyAgency

from reviewboard.reviews.templatetags.reviewtags import has_usable_review_ui
from reviewboard.admin.server import build_server_url
from reviewboard.reviews.ui.base import (FileAttachmentReviewUI,
                                         register_ui,
                                         unregister_ui)
from reviewboard.reviews.ui.image import ImageReviewUI
from reviewboard.testing import TestCase


class InitReviewUI(FileAttachmentReviewUI):
    supported_mimetypes = ['image/jpg']

    def __init__(self, review_request, obj):
        raise Exception


class SandboxReviewUI(FileAttachmentReviewUI):
    supported_mimetypes = ['image/png']

    def is_enabled_for(self, user=None, review_request=None,
                       file_attachment=None, **kwargs):
        raise Exception

    def get_comment_thumbnail(self, comment):
        raise Exception

    def get_comment_link_url(self, comment):
        raise Exception

    def get_comment_link_text(self, comment):
        raise Exception

    def get_extra_context(self, request):
        raise Exception

    def get_js_view_data(self):
        raise Exception


class ConflictFreeReviewUI(FileAttachmentReviewUI):
    supported_mimetypes = ['image/gif']

    def serialize_comment(self, comment):
        raise Exception

    def get_js_model_data(self):
        raise Exception


class SandboxTests(SpyAgency, TestCase):
    """Testing sandboxing extensions."""
    fixtures = ['test_users']

    def setUp(self):
        super(SandboxTests, self).setUp()

        register_ui(InitReviewUI)
        register_ui(SandboxReviewUI)
        register_ui(ConflictFreeReviewUI)

        self.factory = RequestFactory()

        self.user = User.objects.get(username='doc')
        self.review_request = self.create_review_request()

        self.png_file = self.create_file_attachment(self.review_request)
        self.png_file.mimetype = 'image/png'

        self.jpg_file = self.create_file_attachment(self.review_request)
        self.jpg_file.mimetype = 'image/jpg'

        self.gif_file = self.create_file_attachment(self.review_request)
        self.gif_file.mimetype = 'image/gif'

    def tearDown(self):
        super(SandboxTests, self).tearDown()

        unregister_ui(InitReviewUI)
        unregister_ui(SandboxReviewUI)
        unregister_ui(ConflictFreeReviewUI)

    def test_init_review_ui(self):
        """Testing FileAttachmentReviewUI sandboxes __init__"""
        self.spy_on(InitReviewUI.__init__)

        self.jpg_file.review_ui

        self.assertTrue(InitReviewUI.__init__.called)

    def test_is_enabled_for(self):
        """Testing FileAttachmentReviewUI sandboxes is_enabled_for"""
        review_ui = self.png_file.review_ui
        self.spy_on(review_ui.is_enabled_for)

        self.assertFalse(has_usable_review_ui(self.user,
                                              self.review_request,
                                              self.png_file))
        self.assertTrue(review_ui.is_enabled_for.called)

    def test_get_comment_thumbnail(self):
        """Testing FileAttachmentReviewUI sandboxes get_comment_thumbnail"""
        comment = "Comment"

        review = self.create_review(self.review_request, user=self.user)
        file_attachment_comment = review.file_attachment_comments.create(
            file_attachment=self.png_file, text=comment)

        review_ui = file_attachment_comment.review_ui
        self.spy_on(review_ui.get_comment_thumbnail)

        file_attachment_comment.thumbnail

        self.assertTrue(review_ui.get_comment_thumbnail.called)

    def test_get_comment_link_url(self):
        """Testing FileAttachmentReviewUI sandboxes get_comment_link_url"""
        comment = "Comment"

        review = self.create_review(self.review_request, user=self.user)
        file_attachment_comment = review.file_attachment_comments.create(
            file_attachment=self.png_file, text=comment)

        review_ui = file_attachment_comment.review_ui
        self.spy_on(review_ui.get_comment_link_url)

        file_attachment_comment.get_absolute_url()

        self.assertTrue(review_ui.get_comment_link_url.called)

    def test_get_comment_link_text(self):
        """Testing FileAttachmentReviewUI sandboxes get_comment_link_text"""
        comment = "Comment"

        review = self.create_review(self.review_request, user=self.user)
        file_attachment_comment = review.file_attachment_comments.create(
            file_attachment=self.png_file, text=comment)

        review_ui = file_attachment_comment.review_ui
        self.spy_on(review_ui.get_comment_link_text)

        file_attachment_comment.get_link_text()

        self.assertTrue(review_ui.get_comment_link_text.called)

    def test_get_extra_context(self):
        """Testing FileAttachmentReviewUI sandboxes get_extra_context"""
        review_ui = self.png_file.review_ui
        request = self.factory.get('test')
        request.user = self.user

        self.spy_on(review_ui.get_extra_context)

        review_ui.render_to_string(request=request)

        self.assertTrue(review_ui.get_extra_context.called)

    def test_get_js_model_data(self):
        """Testing FileAttachmentReviewUI sandboxes get_js_model_data"""
        review_ui = self.gif_file.review_ui
        request = self.factory.get('test')
        request.user = self.user

        self.spy_on(review_ui.get_js_model_data)

        review_ui.render_to_response(request=request)

        self.assertTrue(review_ui.get_js_model_data.called)

    def test_get_js_view_data(self):
        """Testing FileAttachmentReviewUI sandboxes get_js_view_data"""
        review_ui = self.png_file.review_ui
        request = self.factory.get('test')
        request.user = self.user

        self.spy_on(review_ui.get_js_view_data)

        review_ui.render_to_response(request=request)

        self.assertTrue(review_ui.get_js_view_data.called)

    def test_serialize_comments(self):
        """Testing FileAttachmentReviewUI sandboxes serialize_comments"""
        review_ui = self.gif_file.review_ui

        self.spy_on(review_ui.serialize_comments)

        review_ui.get_comments_json()

        self.assertTrue(review_ui.serialize_comments.called)

    def test_serialize_comment(self):
        """Testing FileAttachmentReviewUI sandboxes serialize_comment"""
        comment = 'comment'

        review_ui = self.gif_file.review_ui
        request = self.factory.get('test')
        request.user = self.user
        review_ui.request = request

        review = self.create_review(self.review_request, user=self.user)
        file_attachment_comment = review.file_attachment_comments.create(
            file_attachment=self.gif_file, text=comment)

        self.spy_on(review_ui.serialize_comment)

        serial_comments = review_ui.serialize_comments(
            comments=[file_attachment_comment])
        self.assertEqual(len(serial_comments), 0)

        self.assertTrue(review_ui.serialize_comment.called)


class ImageReviewUITests(TestCase):
    """Tests for the ImageReviewUI."""

    fixtures = ['test_users']

    def setUp(self):
        self.review_request = self.create_review_request()
        self.attachment = self.create_file_attachment(
            self.review_request)
        self.review = self.create_review(self.review_request)

    def test_get_comment_thumbnail(self):
        """Testing ImageReviewUI.get_comment_thumbnail for an image comment"""
        ui = ImageReviewUI(self.review_request, self.attachment)
        comment = self.create_file_attachment_comment(
            self.review,
            self.attachment,
            extra_fields={
                'x': 0,
                'y': 0,
                'width': 1,
                'height': 1,
            })
        thumbnail = ui.get_comment_thumbnail(comment)

        self.assertEqual(
            thumbnail,
            '<img class="modified-image" src="%s" width="1" height="1"'
            ' alt="%s" />'
            % (build_server_url(crop_image(self.attachment.file, 0, 0, 1, 1)),
               comment.text)
        )

    def test_get_comment_thumbnail_diff(self):
        """Testing ImageReviewUI.get_comment_thumbnail for an image diff
        comment
        """
        diff_attachment = self.create_file_attachment(self.review_request)

        ui = ImageReviewUI(self.review_request, self.attachment)
        ui.set_diff_against(diff_attachment)

        comment = self.create_file_attachment_comment(
            self.review,
            self.attachment,
            diff_attachment,
            extra_fields={
                'x': 0,
                'y': 0,
                'width': 1,
                'height': 1,
            })
        thumbnail = ui.get_comment_thumbnail(comment)

        self.assertEqual(
            thumbnail,
            '<div class="image-review-ui-diff-thumbnail">'
            '<img class="orig-image" src="%s" width="1" height="1" alt="%s" />'
            '<img class="modified-image" src="%s" width="1" height="1"'
            ' alt="%s" />'
            '</div>'
            % (build_server_url(crop_image(diff_attachment.file, 0, 0, 1, 1)),
               comment.text,
               build_server_url(crop_image(self.attachment.file, 0, 0, 1, 1)),
               comment.text)
        )



"""Base support for creating and using integrations.

This module provides a method for getting the integration manager used by
Review Board (:py:func:`get_integration_manager`) and a base class for defining
integrations (:py:class:`Integration`).
"""

from __future__ import unicode_literals

from djblets.integrations.integration import Integration as DjbletsIntegration
from djblets.integrations.manager import IntegrationManager


_integration_manager = None


class Integration(DjbletsIntegration):
    """Base class for an integration.

    Integrations are pluggable components that interface the application with
    a third-party service, notifying the service or triggering actions on
    certain events, or fetching data from the service. They can be registered
    by the consuming application or through extensions.

    Unlike an extension, an integration can contain multiple configurations
    active at one time. This is useful, for instance, when you want multiple,
    distinct configurations for posting messages to different channels on a
    chat service.

    There's one :py:class:`Integration` instance for each class, and it
    typically operates by responding to events and communicating with another
    service, making use of the state stored in one or more
    :py:class:`~reviewboard.integrations.models.IntegrationConfig` instances,
    which it can query. This allows hook registration and other logic to be
    shared across all configurations of an instance.

    Integrations can make use of :ref:`extension-hooks`, binding the lifecycle
    of that hook's registration to the lifecycle of that particular
    integration, making it very easy to tie an integration into any part of
    the application.
    """

    def get_configs(self, local_site):
        """Return configurations matching the given filters.

        This will return all enabled configurations for this integration
        matching the provided ``local_site``.

        The configurations can be filtered down further by the caller, based
        on the settings.

        Args:
            local_site (reviewboard.site.models.LocalSite):
                The Local Site matching any configurations. This should
                correspond to the value used for any repositories, review
                requests, reviews, etc. being used to trigger an operation,
                and should be set based on those objects.

        Returns:
            list of reviewboard.integrations.models.IntegrationConfig:
            A list of enabled integration configurations matching the query.
        """
        return super(Integration, self).get_configs(local_site=local_site)


class GetIntegrationManagerMixin(object):
    """Mixin for supplying an IntegrationManager for classes.

    This is used in any class that needs a :py:meth:`get_integration_manager`
    function.
    """

    @classmethod
    def get_integration_manager(self):
        """Return the IntegrationManager for the class.

        Returns:
            djblets.integrations.manager.IntegrationManager:
            The integration manager used in Review Board.
        """
        return get_integration_manager()


def get_integration_manager():
    """Return the integration manager for Review Board.

    Returns:
        djblets.integrations.manager.IntegrationManager:
        The integration manager used for Review Board.
    """
    global _integration_manager

    if not _integration_manager:
        from reviewboard.integrations.models import IntegrationConfig

        _integration_manager = IntegrationManager(IntegrationConfig)

    return _integration_manager



"""Unit tests for reviewboard.accounts.templatetags.accounts."""

from __future__ import unicode_literals

from django.contrib.auth.models import User
from django.template import Context, Template
from django.test.client import RequestFactory
from django.utils.safestring import SafeText

from reviewboard.testing import TestCase


class UserProfileDisplayNameTests(TestCase):
    """Unit tests for {% user_profile_display_name %}."""

    def test_escapes_name(self):
        """Testing {% user_profile_display_name %} escapes content"""
        user = User.objects.create(
            username='test',
            first_name='"><script>alert("unsafe");</script>',
            last_name='<"',
            email='test@example.com')
        request_user = User.objects.create(username='admin',
                                           email='admin@example.com',
                                           is_superuser=True)
        rendered = self._render_tag(user, request_user)

        self.assertEqual(
            rendered,
            '&quot;&gt;&lt;script&gt;alert(&quot;unsafe&quot;);&lt;/script&gt;'
            ' &lt;&quot;')

    def _render_tag(self, user, request_user):
        """Utility function to render the tag.

        This will render the template for the tag, providing all necessary
        arguments. It will also check the rendered string is of the expected
        type before returning it.

        Args:
            user (django.contrib.auth.models.User):
                The user to pass to the tag.

            request_user (django.contrib.auth.models.User):
                The requesting user.

        Returns:
            django.utils.safestring.SafeText:
            The rendered content.
        """
        request = RequestFactory().get('/')
        request.user = request_user

        t = Template('{% load accounts %}'
                     '{% user_profile_display_name user %}')

        rendered = t.render(Context({
            'request': request,
            'user': user,
        }))

        self.assertIsInstance(rendered, SafeText)

        return rendered



from __future__ import unicode_literals

from django_evolution.mutations import AddField
from django.db import models


MUTATIONS = [
    AddField('Application', 'enabled', models.BooleanField, initial=True),
    AddField('Application', 'original_user', models.ForeignKey, null=True,
             related_model='auth.User'),
]



"""Configuration forms for integrations."""

from __future__ import unicode_literals

from django import forms
from django.utils import six
from django.utils.translation import ugettext_lazy as _
from djblets.forms.fields import ConditionsField
from djblets.integrations.forms import (IntegrationConfigForm as
                                        DjbletsIntegrationConfigForm)

from reviewboard.site.models import LocalSite


class IntegrationConfigForm(DjbletsIntegrationConfigForm):
    """Base class for an integration settings form.

    This makes it easy to provide a basic form for manipulating the settings
    of an integration configuration. It takes care of loading/saving the
    values and prompting the user for a name.

    Integrations should subclass this and provide additional fields that they
    want to display to the user. They must provide a :py:class:`Meta` class
    containing the fieldsets they want to display.

    Attributes:
        limit_to_local_site (reviewboard.site.models.LocalSite):
            The optional LocalSite to limit this configuration to. Any
            configuration-related fields or logic that might need to be bound
            to a LocalSite must make use of this.
    """

    #: A list of fields on the model that should not be saved in settings
    model_fields = DjbletsIntegrationConfigForm.model_fields + ('local_site',)

    #: The fieldset containing basic information on the configuration.
    #:
    #: This is the same as the djblets version, but with the local site field
    #: added in.
    basic_info_fieldset = (None, {
        'fields': ('name', 'enabled', 'local_site'),
        'description': _(
            'Start by giving this configuration a name so you can easily '
            'identify it later. You can also mark this configuration as '
            'enabled or disabled.'
        ),
    })

    local_site = forms.ModelChoiceField(
        label=_('Local Site'),
        queryset=LocalSite.objects.all(),
        required=False)

    def __init__(self, *args, **kwargs):
        """Initialize the form.

        Args:
            limit_to_local_site (reviewboard.site.models.LocalSite, optional):
                The Local Site to limit configurations to. If ``None`` (or not
                provided), the configuration's Local Site (or lack thereof) can
                be specified by the user.

            *args (tuple):
                Positional arguments to pass to the parent form.

            **kwargs (dict):
                Keyword arguments to pass to the parent form.
        """
        local_site = kwargs.pop('limit_to_local_site', None)
        self.limit_to_local_site = local_site

        super(IntegrationConfigForm, self).__init__(*args, **kwargs)

        if local_site:
            self.fields['local_site'].queryset = \
                LocalSite.objects.filter(pk=local_site.pk)

            # Limit LocalSites for all condition fields.
            for field in six.itervalues(self.fields):
                if isinstance(field, ConditionsField):
                    field.choice_kwargs['local_site'] = local_site



#
# reviewboard/admin/urls.py -- URLs for the admin app
#
# Copyright (c) 2008-2009  Christian Hammond
# Copyright (c) 2009  David Trowbridge
#
# Permission is hereby granted, free of charge, to any person obtaining
# a copy of this software and associated documentation files (the
# "Software"), to deal in the Software without restriction, including
# without limitation the rights to use, copy, modify, merge, publish,
# distribute, sublicense, and/or sell copies of the Software, and to
# permit persons to whom the Software is furnished to do so, subject to
# the following conditions:
#
# The above copyright notice and this permission notice shall be included
# in all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
# EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
# MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
# IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY
# CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT,
# TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE
# SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
#

from __future__ import unicode_literals

from django.conf.urls import include, url
from django.contrib import admin
from django.views.generic import RedirectView
from djblets.feedview.views import view_feed

from reviewboard.admin import forms, views


NEWS_FEED = 'https://www.reviewboard.org/news/feed/'

urlpatterns = [
    url(r'^$', views.dashboard),

    url(r'^cache/$', views.cache_stats, name='admin-server-cache'),

    url(r'^db/', include(admin.site.urls)),

    url(r'^integrations/', include('reviewboard.integrations.urls')),

    url(r'^feed/news/$',
        view_feed,
        kwargs={
            'template_name': 'admin/feed.html',
            'url': NEWS_FEED,
        }),

    url(r'^feed/news/rss/$',
        RedirectView.as_view(url=NEWS_FEED, permanent=True)),

    url(r'^log/', include('djblets.log.urls')),

    url(r'^security/$', views.security, name='admin-security-checks'),

    url(r'^settings/', include([
        url(r'^$', RedirectView.as_view(url='general/', permanent=True),
            name='site-settings'),

        url(r'^general/$',
            views.site_settings,
            kwargs={
                'form_class': forms.GeneralSettingsForm,
                'template_name': 'admin/general_settings.html',
            },
            name='settings-general'),

        url(r'^authentication/$',
            views.site_settings,
            kwargs={
                'form_class': forms.AuthenticationSettingsForm,
                'template_name': 'admin/authentication_settings.html',
            },
            name='settings-authentication'),

        url(r'^avatars/$',
            views.site_settings,
            kwargs={
                'form_class': forms.AvatarServicesForm,
                'template_name': 'admin/avatar_settings.html',
            },
            name='settings-avatars'),

        url(r'^email/$',
            views.site_settings,
            kwargs={
                'form_class': forms.EMailSettingsForm,
                'template_name': 'admin/settings.html',
            },
            name='settings-email'),

        url(r'^diffs/$',
            views.site_settings,
            kwargs={
                'form_class': forms.DiffSettingsForm,
                'template_name': 'admin/settings.html',
            },
            name='settings-diffs'),

        url(r'^logging/$',
            views.site_settings,
            kwargs={
                'form_class': forms.LoggingSettingsForm,
                'template_name': 'admin/settings.html',
            },
            name='settings-logging'),

        url(r'^privacy/$',
            views.site_settings,
            kwargs={
                'form_class': forms.PrivacySettingsForm,
                'template_name': 'admin/privacy_settings.html',
            },
            name='settings-privacy'),

        url(r'^ssh/$',
            views.ssh_settings,
            name='settings-ssh'),

        url(r'^storage/$',
            views.site_settings,
            kwargs={
                'form_class': forms.StorageSettingsForm,
                'template_name': 'admin/storage_settings.html',
            },
            name='settings-storage'),

        url(r'^support/$',
            views.site_settings,
            kwargs={
                'form_class': forms.SupportSettingsForm,
                'template_name': 'admin/settings.html',
            },
            name='settings-support'),

        url(r'^search/$',
            views.site_settings,
            kwargs={
                'form_class': forms.SearchSettingsForm,
                'template_name': 'admin/search_settings.html',
            },
            name='settings-search'),
    ])),

    url(r'^widget-activity/$', views.widget_activity),

    url(r'^widget-move/$', views.widget_move),

    url(r'^widget-select/$', views.widget_select),

    url(r'^widget-toggle/$', views.widget_toggle),
]



from __future__ import unicode_literals

from django.contrib.auth.models import User
from haystack import indexes

from reviewboard.search.fields import BooleanField
from reviewboard.search.indexes import BaseSearchIndex


class UserIndex(BaseSearchIndex, indexes.Indexable):
    """A Haystack search index for users."""

    model = User
    local_site_attr = 'local_site'

    username = indexes.CharField(model_attr='username')
    email = indexes.CharField()
    full_name = indexes.CharField()
    url = indexes.CharField(model_attr='get_absolute_url')
    groups = indexes.CharField(indexed=False)
    is_profile_private = BooleanField()

    def index_queryset(self, using=None):
        """Query the list of users for the index.

        All active users will be returned.
        """
        return (
            self.get_model().objects
            .filter(is_active=True)
            .select_related('profile')
            .prefetch_related('review_groups', 'local_site')
        )

    def prepare_groups(self, user):
        """Prepare a user's list of groups for the index.

        Only publicly-accessible groups will be stored in the index.

        Args:
            user (django.contrib.auth.models.User):
                The user being indexed.

        Returns:
            unicode:
            The names of publicly-accessible groups.
        """
        return ','.join(
            review_group.name
            for review_group in user.review_groups.all()
            if not review_group.invite_only
        )

    def prepare_email(self, user):
        """Prepare the email field for a user.

        Args:
            user (django.contrib.auth.models.User):
                The user being indexed.

        Returns:
            unicode:
            The e-mail address.
        """
        profile = user.get_profile(cached_only=True)

        if profile is None or profile.is_private:
            return ''
        else:
            return user.email

    def prepare_full_name(self, user):
        """Prepare the full_name field for a user.

        This field is not included in the index when their profile is set to
        private.

        Args:
            user (django.contrib.auth.models.User):
                The user being indexed.

        Returns:
            unicode:
            The full name of the user or an empty string if their profile is
            private.
        """
        profile = user.get_profile(cached_only=True)

        if profile is None or profile.is_private:
            return ''
        else:
            return user.get_full_name()

    def prepare_is_profile_private(self, user):
        """Prepare the is_profile_private field for a user.

        Args:
            user (django.contrib.auth.models.User):
                The user being indexed.

        Returns:
            bool:
            Whther or not the profile is private.
        """
        profile = user.get_profile(cached_only=True)
        return profile is None or profile.is_private



"""Signal handlers."""

from __future__ import unicode_literals

from django.contrib.auth.models import User

from reviewboard.oauth.models import Application


def on_users_changed(instance, action, pk_set, reverse, **kwargs):
    """Handle the users of a Local Site changing.

    This method ensures that any
    :py:class:`Applications <reviewboard.oauth.models.Application>` owned by
    users removed from a a :py:class:`~reviewboard.site.models.LocalSite` will
    be re-assigned to an administrator on that Local Site and disabled so the
    client secret can be changed.

    Args:
        instance (django.contrib.auth.models.User or
                  reviewboard.reviews.models.review_group.Group):
            The model that changed.

        action (unicode):
            The change action on the Local Site.

        pk_set (list of int):
            The primary keys of the objects changed.

        reverse (bool):
            Whether or not the relation or the reverse relation is changing.

        **kwargs (dict):
            Ignored arguments from the signal.
    """
    users = None

    # When reverse is True, `instance` will be a user that was changed (i.e.,
    # the signal triggered from user.local_sites.add(site)). Otherwise,
    # `instance` will be the `local_site` that changed and `pk_set `will be
    # the list of user primary keys that were added/removed.
    if action == 'post_remove':
        if reverse:
            users = [instance]
        else:
            users = list(User.objects.filter(pk__in=pk_set))
    elif action == 'pre_clear':
        if reverse:
            users = [instance]
        else:
            # We have to grab the list of associated users in the pre_clear
            # phase because pk_set is always empty for pre_ and post_clear.
            users = list(instance.users.all())

    if not users:
        return

    applications = list(
        Application.objects
        .filter(user__in=users,
                local_site__isnull=False)
        .prefetch_related('local_site__admins')
    )

    if not applications:
        return

    users_by_pk = {
        user.pk: user
        for user in users
    }

    for application in applications:
        user = users_by_pk[application.user_id]

        if not application.local_site.is_accessible_by(user):
            # The user who owns this application no longer has access to the
            # Local Site. We must disable the application and
            application.enabled = False
            application.user = application.local_site.admins.first()
            application.original_user = user
            application.save(update_fields=[
                'enabled',
                'original_user',
                'user',
            ])



"""Unit tests for reviewboard.reviews.views.DownloadRawDiffView."""

from __future__ import unicode_literals

from reviewboard.testing import TestCase


class DownloadRawDiffViewTests(TestCase):
    """Unit tests for reviewboard.reviews.views.DownloadRawDiffView."""

    fixtures = ['test_users', 'test_scmtools']

    # Bug #3384
    def test_sends_correct_content_disposition(self):
        """Testing DownloadRawDiffView sends correct Content-Disposition"""
        review_request = self.create_review_request(create_repository=True,
                                                    publish=True)

        self.create_diffset(review_request=review_request)

        response = self.client.get('/r/%d/diff/raw/' % review_request.pk)
        self.assertEqual(response.status_code, 200)
        self.assertEqual(response['Content-Disposition'],
                         'attachment; filename=diffset')

    # Bug #3704
    def test_normalize_commas_in_filename(self):
        """Testing DownloadRawDiffView removes commas in filename"""
        review_request = self.create_review_request(create_repository=True,
                                                    publish=True)

        # Create a diffset with a comma in its name.
        self.create_diffset(review_request=review_request, name='test, comma')

        response = self.client.get('/r/%d/diff/raw/' % review_request.pk)
        content_disposition = response['Content-Disposition']
        filename = content_disposition[len('attachment; filename='):]
        self.assertFalse(',' in filename)
